import copy
from typing import Any, Tuple

import numpy as np
import pytest
from neuraxle.base import BaseStep
from neuraxle.steps.column_transformer import ColumnTransformer


class MultiplyBy2(BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        self.fitted_data = []

    def fit(self, data_inputs, expected_outputs=None) -> 'BaseStep':
        self._add_fitted_data(data_inputs, expected_outputs)
        return self

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        self._add_fitted_data(data_inputs, expected_outputs)
        return self, self.transform(data_inputs)

    def _add_fitted_data(self, data_inputs, expected_outputs):
        self.fitted_data.append((copy.deepcopy(data_inputs), copy.deepcopy(expected_outputs)))

    def transform(self, data_inputs):
        return (2 * np.array(data_inputs)).tolist()


class ColumnChooserTestCase:
    def __init__(
            self,
            data_inputs,
            expected_processed_outputs,
            column_transformer_tuple_list,
            n_dimension,
            expected_step_key=None,
            expected_fitted_data=None,
            expected_outputs=None
    ):
        self.n_dimension = n_dimension
        self.expected_step_key = expected_step_key
        self.data_inputs = data_inputs
        self.expected_outputs = expected_outputs
        self.expected_fitted_data = expected_fitted_data
        self.expected_processed_outputs = expected_processed_outputs
        self.column_transformer_tuple_list = column_transformer_tuple_list


# 2d
test_case_index_int_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [2],
        [22],
        [42]
    ]),
    expected_fitted_data=[(
        [[1], [11], [21]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='1_MultiplyBy2',
    column_transformer_tuple_list=[
        (1, MultiplyBy2())
    ],
    n_dimension=2
)

test_case_index_start_end_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [0, 2],
        [20, 22],
        [40, 42]
    ]),
    expected_fitted_data=[(
        [[0, 1], [10, 11], [20, 21]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='slice(0, 2, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(0, 2), MultiplyBy2())
    ],
    n_dimension=2
)

test_case_index_range_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [0, 2],
        [20, 22],
        [40, 42]
    ]),
    expected_fitted_data=[(
        [[0, 1], [10, 11], [20, 21]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='range(0, 2)_MultiplyBy2',
    column_transformer_tuple_list=[
        (range(2), MultiplyBy2())
    ],
    n_dimension=2
)

test_case_index_start_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [2, 4, 6],
        [22, 24, 26],
        [42, 44, 46]
    ]),
    expected_fitted_data=[(
        [[1, 2, 3], [11, 12, 13], [21, 22, 23]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='slice(1, None, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(1, None), MultiplyBy2())
    ],
    n_dimension=2
)

test_case_index_end_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [0, 2],
        [20, 22],
        [40, 42]
    ]),
    expected_fitted_data=[(
        [[0, 1], [10, 11], [20, 21]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='slice(None, 2, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(None, 2), MultiplyBy2())
    ],
    n_dimension=2
)

test_case_index_last_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [0, 2, 4],
        [20, 22, 24],
        [40, 42, 44]
    ]),
    expected_fitted_data=[
        (
            [[0, 1, 2], [10, 11, 12], [20, 21, 22]],
            [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
        )
    ],
    expected_step_key='slice(None, -1, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(None, -1), MultiplyBy2())
    ],
    n_dimension=2
)

test_case_list_of_columns_2d = ColumnChooserTestCase(
    data_inputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_outputs=np.array([
        [0, 1, 2, 3],
        [10, 11, 12, 13],
        [20, 21, 22, 23]
    ]),
    expected_processed_outputs=np.array([
        [0, 4],
        [20, 24],
        [40, 44]
    ]),
    expected_fitted_data=[(
        [[0, 2], [10, 12], [20, 22]],
        [[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]]
    )],
    expected_step_key='[0, 2]_MultiplyBy2',
    column_transformer_tuple_list=[
        ([0, 2], MultiplyBy2())
    ],
    n_dimension=2
)

# 3d
test_case_index_int = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[2]],
        [[22]],
        [[42]]
    ]),
    expected_fitted_data=[
        ([np.array([[1]]), np.array([[11]]), np.array([[21]])],
         [np.array([[0, 1, 2, 3]]),
          np.array([[10, 11, 12, 13]]),
          np.array([[20, 21, 22, 23]])]
         )
    ],
    expected_step_key='1_MultiplyBy2',
    column_transformer_tuple_list=[
        (1, MultiplyBy2())
    ],
    n_dimension=3
)

# test_case_index_int = ColumnChooserTestCase(
#     data_inputs=np.array([[
#         [0, 1, 2, 3],
#         [10, 11, 12, 13],
#         [20, 21, 22, 23]
#     ]]),
#     expected_outputs=np.array([[
#         [0, 1, 2, 3],
#         [10, 11, 12, 13],
#         [20, 21, 22, 23]
#     ]]),
#     expected_processed_outputs=np.array([[
#         [2],
#         [22],
#         [42]
#     ]]),
#     expected_fitted_data=[(
#         [np.array([[1], [11], [21]])],
#         [np.array([[0, 1, 2, 3], [10, 11, 12, 13], [20, 21, 22, 23]])]
#     )],
#     expected_step_key='1_MultiplyBy2',
#     column_transformer_tuple_list=[
#         (1, MultiplyBy2())
#     ],
#     n_dimension=3
# )

test_case_index_start_end = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[0, 2]],
        [[20, 22]],
        [[40, 42]]
    ]),
    expected_fitted_data=[(
        [np.array([[0, 1]]), np.array([[10, 11]]), np.array([[20, 21]])],
        [np.array([[0, 1, 2, 3]]),
         np.array([[10, 11, 12, 13]]),
         np.array([[20, 21, 22, 23]])]
    )],
    expected_step_key='slice(0, 2, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(0, 2), MultiplyBy2())
    ],
    n_dimension=3
)

test_case_index_range = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[0, 2]],
        [[20, 22]],
        [[40, 42]]
    ]),
    expected_fitted_data=[(
        [np.array([[0, 1]]), np.array([[10, 11]]), np.array([[20, 21]])],
        [np.array([[0, 1, 2, 3]]),
         np.array([[10, 11, 12, 13]]),
         np.array([[20, 21, 22, 23]])]
    )],
    expected_step_key='range(0, 2)_MultiplyBy2',
    column_transformer_tuple_list=[
        (range(2), MultiplyBy2())
    ],
    n_dimension=3
)

test_case_index_start = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[2, 4, 6]],
        [[22, 24, 26]],
        [[42, 44, 46]]
    ]),
    expected_fitted_data=[(
        [np.array([[1, 2, 3]]),
         np.array([[11, 12, 13]]),
         np.array([[21, 22, 23]])],
        [np.array([[0, 1, 2, 3]]),
         np.array([[10, 11, 12, 13]]),
         np.array([[20, 21, 22, 23]])]
    )],
    expected_step_key='slice(1, None, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(1, None), MultiplyBy2())
    ],
    n_dimension=3
)

test_case_index_end = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[0, 2]],
        [[20, 22]],
        [[40, 42]]
    ]),
    expected_fitted_data=[(
        [np.array([[0, 1]]),
         np.array([[10, 11]]),
         np.array([[20, 21]])],
        [np.array([[0, 1, 2, 3]]),
         np.array([[10, 11, 12, 13]]),
         np.array([[20, 21, 22, 23]])]
    )],
    expected_step_key='slice(None, 2, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(None, 2), MultiplyBy2())
    ],
    n_dimension=3
)

test_case_index_last = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[0, 2, 4]],
        [[20, 22, 24]],
        [[40, 42, 44]]
    ]),
    expected_fitted_data=[
        (
            [np.array([[0, 1, 2]]),
             np.array([[10, 11, 12]]),
             np.array([[20, 21, 22]])],
            [np.array([[0, 1, 2, 3]]),
             np.array([[10, 11, 12, 13]]),
             np.array([[20, 21, 22, 23]])]
        )
    ],
    expected_step_key='slice(None, -1, None)_MultiplyBy2',
    column_transformer_tuple_list=[
        (slice(None, -1), MultiplyBy2())
    ],
    n_dimension=3
)

test_case_list_of_columns = ColumnChooserTestCase(
    data_inputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_outputs=np.array([
        [[0, 1, 2, 3]],
        [[10, 11, 12, 13]],
        [[20, 21, 22, 23]]
    ]),
    expected_processed_outputs=np.array([
        [[0, 4]],
        [[20, 24]],
        [[40, 44]]
    ]),
    expected_fitted_data=[(
        [np.array([[0, 2]]),
         np.array([[10, 12]]),
         np.array([[20, 22]])],
        [np.array([[0, 1, 2, 3]]),
         np.array([[10, 11, 12, 13]]),
         np.array([[20, 21, 22, 23]])]
    )],
    expected_step_key='[0, 2]_MultiplyBy2',
    column_transformer_tuple_list=[
        ([0, 2], MultiplyBy2())
    ],
    n_dimension=3
)


@pytest.mark.parametrize("test_case", [
    copy.deepcopy(test_case_index_int),
    copy.deepcopy(test_case_index_start_end),
    copy.deepcopy(test_case_index_start),
    copy.deepcopy(test_case_index_range),
    copy.deepcopy(test_case_index_end),
    copy.deepcopy(test_case_index_last),
    copy.deepcopy(test_case_list_of_columns),
    copy.deepcopy(test_case_index_int_2d),
    copy.deepcopy(test_case_index_start_end_2d),
    copy.deepcopy(test_case_index_start_2d),
    copy.deepcopy(test_case_index_range_2d),
    copy.deepcopy(test_case_index_end_2d),
    copy.deepcopy(test_case_index_last_2d),
    copy.deepcopy(test_case_list_of_columns_2d)
])
def test_column_transformer_transform_should_support_indexes(test_case: ColumnChooserTestCase):
    data_inputs = test_case.data_inputs
    column_transformer = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    outputs = column_transformer.transform(data_inputs)

    assert np.array_equal(outputs, test_case.expected_processed_outputs)


@pytest.mark.parametrize("test_case", [
    copy.deepcopy(test_case_index_int),
    copy.deepcopy(test_case_index_start_end),
    copy.deepcopy(test_case_index_start),
    copy.deepcopy(test_case_index_range),
    copy.deepcopy(test_case_index_end),
    copy.deepcopy(test_case_index_last),
    copy.deepcopy(test_case_list_of_columns),
    copy.deepcopy(test_case_index_int_2d),
    copy.deepcopy(test_case_index_start_end_2d),
    copy.deepcopy(test_case_index_start_2d),
    copy.deepcopy(test_case_index_range_2d),
    copy.deepcopy(test_case_index_end_2d),
    copy.deepcopy(test_case_index_last_2d),
    copy.deepcopy(test_case_list_of_columns_2d)
])
def test_column_transformer_fit_transform_should_support_indexes(test_case: ColumnChooserTestCase):
    data_inputs = test_case.data_inputs
    expected_outputs = test_case.expected_outputs
    p = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, test_case.expected_processed_outputs)
    actual_fitted_data = p[test_case.expected_step_key]['MultiplyBy2'].fitted_data
    expected_fitted_data = test_case.expected_fitted_data
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)


@pytest.mark.parametrize("test_case", [
    copy.deepcopy(test_case_index_int),
    copy.deepcopy(test_case_index_start_end),
    copy.deepcopy(test_case_index_start),
    copy.deepcopy(test_case_index_range),
    copy.deepcopy(test_case_index_end),
    copy.deepcopy(test_case_index_last),
    copy.deepcopy(test_case_list_of_columns),
    copy.deepcopy(test_case_index_int_2d),
    copy.deepcopy(test_case_index_start_end_2d),
    copy.deepcopy(test_case_index_start_2d),
    copy.deepcopy(test_case_index_range_2d),
    copy.deepcopy(test_case_index_end_2d),
    copy.deepcopy(test_case_index_last_2d),
    copy.deepcopy(test_case_list_of_columns_2d)
])
def test_column_transformer_fit_should_support_indexes(test_case: ColumnChooserTestCase):
    data_inputs = test_case.data_inputs
    p = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    p = p.fit(data_inputs, test_case.expected_outputs)

    actual_fitted_data = p[test_case.expected_step_key]['MultiplyBy2'].fitted_data
    expected_fitted_data = test_case.expected_fitted_data
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)


def test_column_transformer_fit_should_support_multiple_tuples():
    # Given
    test_case = ColumnChooserTestCase(
        data_inputs=np.array([
            [[1, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_outputs=np.array([
            [[0, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_processed_outputs=np.array([
            [[2, 2, 2, 3]],
            [[20, 22, 12, 13]],
            [[40, 42, 44, 46]]
        ]),
        column_transformer_tuple_list=[
            (slice(0, 2), MultiplyBy2()),
            (2, MultiplyBy2())
        ],
        n_dimension=3
    )
    data_inputs = test_case.data_inputs
    p = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    # When
    p = p.fit(data_inputs, test_case.expected_outputs)

    # Then
    actual_fitted_data = p['2_MultiplyBy2']['MultiplyBy2'].fitted_data
    expected_fitted_data = [
        ([np.array([[2]]), np.array([[12]]), np.array([[22]])],
         [np.array([[0, 1, 2, 3]]), np.array([[10, 11, 12, 13]]), np.array([[20, 21, 22, 23]])])
    ]
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)

    actual_fitted_data = p['slice(0, 2, None)_MultiplyBy2']['MultiplyBy2'].fitted_data
    expected_fitted_data = [([np.array([[1, 1]]), np.array([[10, 11]]), np.array([[20, 21]])],
                              [np.array([[0, 1, 2, 3]]), np.array([[10, 11, 12, 13]]), np.array([[20, 21, 22, 23]])]
                            )]
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)


def test_column_transformer_fit_transform_should_support_multiple_tuples():
    # Given
    test_case = ColumnChooserTestCase(
        data_inputs=np.array([
            [[1, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_outputs=np.array([
            [[0, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_processed_outputs=np.array([
            [[2, 2, 4]],
            [[20, 22, 24]],
            [[40, 42, 44]]
        ]),
        column_transformer_tuple_list=[
            (slice(0, 2), MultiplyBy2()),
            (2, MultiplyBy2())
        ],
        n_dimension=3
    )
    data_inputs = test_case.data_inputs
    p = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    # When
    p, outputs = p.fit_transform(data_inputs, test_case.expected_outputs)

    # Then
    assert np.array_equal(test_case.expected_processed_outputs, outputs)
    actual_fitted_data = p['2_MultiplyBy2']['MultiplyBy2'].fitted_data
    expected_fitted_data = [
        ([np.array([[2]]), np.array([[12]]), np.array([[22]])],
         [np.array([[0, 1, 2, 3]]), np.array([[10, 11, 12, 13]]), np.array([[20, 21, 22, 23]])])
                            ]
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)

    actual_fitted_data = p['slice(0, 2, None)_MultiplyBy2']['MultiplyBy2'].fitted_data
    expected_fitted_data = [
        ([np.array([[1, 1]]), np.array([[10, 11]]), np.array([[20, 21]])],
         [np.array([[0, 1, 2, 3]]), np.array([[10, 11, 12, 13]]), np.array([[20, 21, 22, 23]])])
        ]
    assert_data_fitted_properly(actual_fitted_data, expected_fitted_data)


def test_column_transformer_transform_should_support_multiple_tuples():
    # Given
    test_case = ColumnChooserTestCase(
        data_inputs=np.array([
            [[1, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_outputs=np.array([
            [[0, 1, 2, 3]],
            [[10, 11, 12, 13]],
            [[20, 21, 22, 23]]
        ]),
        expected_processed_outputs=np.array([
            [[2, 2, 4]],
            [[20, 22, 24]],
            [[40, 42, 44]]
        ]),
        column_transformer_tuple_list=[
            (slice(0, 2), MultiplyBy2()),
            (2, MultiplyBy2())
        ],
        n_dimension=3
    )
    data_inputs = test_case.data_inputs
    p = ColumnTransformer(test_case.column_transformer_tuple_list, test_case.n_dimension)

    # When
    outputs = p.transform(data_inputs)

    # Then
    assert np.array_equal(test_case.expected_processed_outputs, outputs)


def assert_data_fitted_properly(actual_fitted_data, expected_fitted_data):
    for actual_fitted, expected_fitted in zip(actual_fitted_data, expected_fitted_data):
        assert np.array_equal(actual_fitted[0], expected_fitted[0])
        assert np.array_equal(actual_fitted[1], expected_fitted[1])



from typing import Any, Tuple

import pytest
from neuraxle.base import BaseService, BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import Identity, MetaStep, NamedStepsList, _HasChildrenMixin
from neuraxle.hyperparams.space import RecursiveDict
from neuraxle.pipeline import Pipeline

from testing.test_pipeline import SomeStep


class SomePipeline(Pipeline):
    def __init__(self, steps: NamedStepsList):
        Pipeline.__init__(self, steps)
        self.teared_down = False

    def teardown(self) -> 'BaseStep':
        self.teared_down = True
        return Pipeline.teardown(self)


class SomeException(BaseStep):
    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        raise Exception()

    def fit(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        raise Exception()

    def transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        raise Exception()


class SomeStepSetup(SomeStep):
    def __init__(self):
        SomeStep.__init__(self)
        self.called_with = None


def test_fit_transform_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])

    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.fit_transform([1], [1])

    assert p.is_initialized
    assert step_setup.is_initialized


def test_transform_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])
    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.transform([1])

    assert p.is_initialized
    assert step_setup.is_initialized


def test_fit_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])
    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.fit([1], [1])

    assert p.is_initialized
    assert step_setup.is_initialized


class SomeService(BaseService):
    pass


@pytest.mark.parametrize('base_service', [
    Identity(),
    MetaStep(Identity()),
    SomePipeline([SomeStepSetup()])
])
def test_that_steps_are_setuppeable(base_service: BaseService, tmpdir):
    assert not base_service.is_initialized
    _verify_subservices_initialization(base_service, False)
    base_service.setup(CX(tmpdir))
    _verify_subservices_initialization(base_service, True)
    base_service.teardown()
    _verify_subservices_initialization(base_service, False)


def _verify_subservices_initialization(sub_service, is_initialized: bool):
    assert sub_service.is_initialized == is_initialized
    if isinstance(sub_service, _HasChildrenMixin):
        for child in sub_service.get_children():
            _verify_subservices_initialization(child, is_initialized)


@pytest.mark.parametrize('base_service', [
    Identity(),
    MetaStep(Identity()),
    SomePipeline([SomeStepSetup()]),
    CX(),
    CX().set_service_locator({
        Identity: Identity(),
        SomeService: SomeService()
    }),
    CX().set_service_locator({
        Pipeline: Pipeline([SomeStepSetup()])
    })
])
def test_that_steps_are_applyable_with_name(base_service: BaseService, tmpdir):

    names = base_service.apply(lambda self: RecursiveDict({"name": self.get_name()}))

    _verify_subservices_names(base_service, names)


def _verify_subservices_names(sub_service, sub_service_name: RecursiveDict):
    assert sub_service.name == sub_service_name["name"], f"Not equal: {sub_service.name} != {sub_service_name['name']}."
    if isinstance(sub_service, _HasChildrenMixin):
        for child in sub_service.get_children():
            _verify_subservices_names(child, sub_service_name[child.name])



"""
Tests for Hyperparameters Distribution Spaces
=============================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import copy
import pytest
from collections import OrderedDict
import scipy

from neuraxle.hyperparams.distributions import *
from neuraxle.hyperparams.scipy_distributions import *
from neuraxle.hyperparams.space import HyperparameterSpace, HyperparameterSamples, RecursiveDict

hyperparams_flat_and_dict_pairs = [
    # Pair 1:
    ({
         "a__learning_rate": 7
     },
     {
         "a": {
             "learning_rate": 7
         }
     }),
    # Pair 2:
    ({
         "b__a__learning_rate": 7,
         "b__learning_rate": 9
     },
     {
         "b": {
             "a": {
                 "learning_rate": 7
             },
             "learning_rate": 9
         }
     }),
]


@pytest.mark.parametrize("class_to_test", [RecursiveDict, HyperparameterSamples])
@pytest.mark.parametrize("flat,expected_dic", hyperparams_flat_and_dict_pairs)
def test_flat_to_dict_hyperparams(flat: dict, expected_dic: dict, class_to_test):
    from_flat_dic = class_to_test(flat)
    from_nested_dic = class_to_test(expected_dic)

    assert from_flat_dic == from_nested_dic
    assert from_flat_dic.to_flat_dict() == flat
    assert from_nested_dic.to_flat_dict() == flat
    assert from_nested_dic.to_nested_dict() == expected_dic
    assert from_flat_dic.to_nested_dict() == expected_dic


HYPE_SPACE = HyperparameterSpace(OrderedDict({
    "a__test": Boolean(),
    "a__lr": Choice([0, 1, False, "Test"]),
    "a__b__c": PriorityChoice([0, 1, False, "Test"]),
    "a__b__q": Quantized(Uniform(-10, 10)),
    "d__param": RandInt(-10, 10),
    "d__u": Uniform(-10, 10),
    "e__other": LogUniform(0.001, 10),
    "e__alpha": Normal(0.0, 1.0),
    "e__f__g": LogNormal(0.0, 2.0),
    "p__could_also_be_as_fixed": FixedHyperparameter("also hey"),
    "scipy__poisson": Poisson(1.0, 2.0),
    "scipy__gaussian": Gaussian(-1, 1),
    "scipy__scipy__gaussian": scipy.stats.randint(0, 10)
}))


def test_hyperparams_space_rvs_outputs_samples():
    space = copy.deepcopy(HYPE_SPACE)

    samples = space.rvs()

    assert isinstance(samples, HyperparameterSamples)
    assert len(samples) == len(space)
    for k, v in samples.iter_flat():
        assert k in space
        assert not isinstance(v, HyperparameterDistribution)


@pytest.mark.parametrize("hd", list(HYPE_SPACE.to_flat_dict().values()))
def test_hyperparams_space_rvs_outputs_in_range(hd: HyperparameterDistribution):
    for _ in range(20):

        sample = hd.rvs()

        assert sample in hd



"""
Tests for Metaopt
=============================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import math

import numpy as np
import pytest
from neuraxle.metaopt.validation import (
    AnchoredWalkForwardTimeSeriesCrossValidationSplitter,
    WalkForwardTimeSeriesCrossValidationSplitter)

classic_walforward_parameters = [
    # (training_size, validation_size, padding_between_training_and_validation, drop_remainder)
    # Pair 1:
    (9, 3, 1, False),
    # Pair 2:
    (9, 3, 1, True),
    # Pair 3:
    (9, 3, 2, False),
    # Pair 4:
    (8, 3, 2, False),
    # Pair 4:
    (5, 2, 3, True),
    # Pair 5 (Default Parameters):
    (2, None, 0, False),
]


@pytest.mark.parametrize(
    "training_window_size, validation_window_size, padding_between_training_and_validation, drop_remainder",
    classic_walforward_parameters)
def test_classic_walkforward_crossvalidation_split(
    training_window_size: int,
    validation_window_size: int,
    padding_between_training_and_validation: int,
    drop_remainder: bool
):
    # Arrange
    # set random seed
    np.random.seed(10)

    # Defines the data shape.
    batch_size = 7
    time_series_size = 20
    features_size = 2

    validation_window_size_temp = validation_window_size

    if validation_window_size is None:
        validation_window_size_temp = training_window_size

    if drop_remainder:
        # We have one less number of fold if we drop remainder.
        number_of_fold = math.floor(
            (time_series_size - training_window_size - padding_between_training_and_validation) /
            validation_window_size_temp
        )
    else:
        number_of_fold = math.ceil(
            (time_series_size - training_window_size - padding_between_training_and_validation) /
            validation_window_size_temp
        )

    # Calculate the size of the remainder
    remainder_size = (time_series_size - training_window_size - padding_between_training_and_validation) % \
        validation_window_size_temp
    if remainder_size == 0:
        # Remainder of 0 means the data perfecftly fits in the number of fold and remainder should window_size instead.
        remainder_size = validation_window_size_temp

    # Initialize the inputs.
    data_inputs = np.random.randint(low=0, high=1, size=(batch_size, time_series_size, features_size)).astype(float)
    expected_outputs = np.random.randint(low=0, high=1, size=(batch_size, time_series_size, features_size)) \
        .astype(float)

    # Initialize the class to test.
    step = WalkForwardTimeSeriesCrossValidationSplitter(
        validation_window_size=validation_window_size,
        training_window_size=training_window_size,
        padding_between_training_and_validation=padding_between_training_and_validation,
        drop_remainder=drop_remainder
    )

    # Act
    train_data_inputs, train_expected_outputs, _, validation_data_inputs, validation_expected_outputs, _ = \
        step.split(data_inputs, None, expected_outputs)

    # Assert
    assert len(train_data_inputs) == number_of_fold
    assert len(train_expected_outputs) == number_of_fold
    assert len(validation_data_inputs) == number_of_fold
    assert len(validation_expected_outputs) == number_of_fold

    assert train_data_inputs[0].shape == (batch_size, training_window_size, features_size)
    assert train_expected_outputs[0].shape == (batch_size, training_window_size, features_size)
    assert validation_data_inputs[0].shape == (batch_size, validation_window_size_temp, features_size)
    assert validation_expected_outputs[0].shape == (batch_size, validation_window_size_temp, features_size)

    assert train_data_inputs[1].shape == (batch_size, training_window_size, features_size)
    assert train_expected_outputs[1].shape == (batch_size, training_window_size, features_size)
    assert validation_data_inputs[1].shape == (batch_size, validation_window_size_temp, features_size)
    assert validation_expected_outputs[1].shape == (batch_size, validation_window_size_temp, features_size)

    if drop_remainder:
        assert train_data_inputs[-1].shape == (
            batch_size, training_window_size, features_size)
        assert train_expected_outputs[-1].shape == (
            batch_size, training_window_size, features_size)
        assert validation_data_inputs[-1].shape == (batch_size, validation_window_size_temp, features_size)
        assert validation_expected_outputs[-1].shape == (batch_size, validation_window_size_temp, features_size)
    else:
        assert train_data_inputs[-1].shape == (
            batch_size, training_window_size, features_size)
        assert train_expected_outputs[-1].shape == (
            batch_size, training_window_size, features_size)
        assert validation_data_inputs[-1].shape == (batch_size, remainder_size, features_size)
        assert validation_expected_outputs[-1].shape == (batch_size, remainder_size, features_size)


anchored_walforward_parameters = [
    # (minimum_training_size, validation_window_size, padding_between_training_and_validation, drop_remainder)
    # Pair 1:
    (9, 3, 1, False),
    # Pair 2:
    (9, 3, 1, True),
    # Pair 3:
    (9, 3, 2, False),
    # Pair 4:
    (8, 3, 2, False),
    # Pair 4:
    (5, 2, 3, True),
    # Pair 5 (Default Parameters):
    (2, None, 0, False),
]


@pytest.mark.parametrize(
    "minimum_training_size, validation_window_size, padding_between_training_and_validation, drop_remainder",
    anchored_walforward_parameters)
def test_anchored_walkforward_crossvalidation_split(
        minimum_training_size: int,
        validation_window_size: int,
        padding_between_training_and_validation: int,
        drop_remainder: bool
):
    # Arrange
    # set random seed
    np.random.seed(10)

    # Defines the data shape.
    batch_size = 7
    time_series_size = 20
    features_size = 2

    validation_window_size_temp = validation_window_size

    if validation_window_size is None:
        validation_window_size_temp = minimum_training_size

    if drop_remainder:
        # We have one less number of fold if we drop remainder.
        number_of_fold = math.floor(
            (time_series_size - minimum_training_size - padding_between_training_and_validation) /
            validation_window_size_temp
        )
    else:
        number_of_fold = math.ceil(
            (time_series_size - minimum_training_size - padding_between_training_and_validation) /
            validation_window_size_temp
        )

    # Calculate the size of the remainder
    remainder_size = (time_series_size - minimum_training_size - padding_between_training_and_validation) % \
        validation_window_size_temp
    if remainder_size == 0:
        # Remainder of 0 means the data perfectly fits in the number of fold and remainder should window_size instead.
        remainder_size = validation_window_size_temp

    # Initialize the inputs.
    data_inputs = np.random.randint(low=0, high=1, size=(batch_size, time_series_size, features_size)).astype(float)
    expected_outputs = np.random.randint(low=0, high=1, size=(batch_size, time_series_size, features_size)) \
        .astype(float)

    # Initialize the class to test.
    step = AnchoredWalkForwardTimeSeriesCrossValidationSplitter(
        validation_window_size=validation_window_size,
        minimum_training_size=minimum_training_size,
        padding_between_training_and_validation=padding_between_training_and_validation,
        drop_remainder=drop_remainder
    )

    # Act
    train_data_inputs, train_expected_outputs, _, validation_data_inputs, validation_expected_outputs, _ = \
        step.split(data_inputs, None, expected_outputs)

    # Assert
    assert len(train_data_inputs) == number_of_fold
    assert len(train_expected_outputs) == number_of_fold
    assert len(validation_data_inputs) == number_of_fold
    assert len(validation_expected_outputs) == number_of_fold

    assert train_data_inputs[0].shape == (batch_size, minimum_training_size, features_size)
    assert train_expected_outputs[0].shape == (batch_size, minimum_training_size, features_size)
    assert validation_data_inputs[0].shape == (batch_size, validation_window_size_temp, features_size)
    assert validation_expected_outputs[0].shape == (batch_size, validation_window_size_temp, features_size)

    assert train_data_inputs[1].shape == (batch_size, minimum_training_size + validation_window_size_temp,
                                          features_size)
    assert train_expected_outputs[1].shape == (batch_size, minimum_training_size + validation_window_size_temp,
                                               features_size)
    assert validation_data_inputs[1].shape == (batch_size, validation_window_size_temp, features_size)
    assert validation_expected_outputs[1].shape == (batch_size, validation_window_size_temp, features_size)

    if drop_remainder:
        assert train_data_inputs[-1].shape == (
            batch_size, minimum_training_size + (number_of_fold - 1) * validation_window_size_temp, features_size)
        assert train_expected_outputs[-1].shape == (
            batch_size, minimum_training_size + (number_of_fold - 1) * validation_window_size_temp, features_size)
        assert validation_data_inputs[-1].shape == (batch_size, validation_window_size_temp, features_size)
        assert validation_expected_outputs[-1].shape == (batch_size, validation_window_size_temp, features_size)
    else:
        assert train_data_inputs[-1].shape == (
            batch_size, time_series_size - remainder_size - padding_between_training_and_validation, features_size)
        assert train_expected_outputs[-1].shape == (
            batch_size, time_series_size - remainder_size - padding_between_training_and_validation, features_size)
        assert validation_data_inputs[-1].shape == (batch_size, remainder_size, features_size)
        assert validation_expected_outputs[-1].shape == (batch_size, remainder_size, features_size)



import logging
import os
import shutil
from typing import Set

import numpy as np
import pytest
from neuraxle.base import BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import HandleOnlyMixin, Identity, TrialStatus
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.distributions import FixedHyperparameter
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.logging.logging import NEURAXLE_LOGGER_NAME, NeuraxleLogger
from neuraxle.metaopt.auto_ml import AutoML
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT, AutoMLContext,
                                           ProjectDataclass, ScopedLocation)
from neuraxle.metaopt.validation import ValidationSplitter, RandomSearchSampler
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import AddN, MultiplyByN, NumpyReshape
from sklearn.metrics import mean_squared_error


class FitTransformCounterLoggingStep(HandleOnlyMixin, BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        HandleOnlyMixin.__init__(self)
        self.logging_call_counter = 0

    def _fit_data_container(self, data_container: DACT, context: CX) -> BaseStep:
        self._log(context, "fit")
        return self

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        self._log(context, "transform")
        return data_container

    def _fit_transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        self._log(context, "fit_transform")
        return data_container

    def _log(self, context, name):
        context.logger.warning(f"{name} call - logging call # {self.logging_call_counter}")
        self.logging_call_counter += 1


def test_context_logger_log_file(tmpdir):
    # Setup
    file_path = os.path.join(tmpdir, "test.log")
    if os.path.exists(file_path):
        os.remove(file_path)
    try:

        # Given
        cx = CX(tmpdir)
        cx.logger.with_file_handler(file_path)
        pipeline = Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
            FitTransformCounterLoggingStep()
        ])

        # When
        dact = DACT(
            data_inputs=np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        )
        pipeline.handle_fit(dact, cx)

        # Then
        log = ''.join(cx.logger.read_log_file())
        assert "fit call - logging call # 0" in log

    # Teardown
    finally:
        cx.logger.without_file_handler()
        os.remove(file_path)


class TestTrialLogger:

    def test_logger_automl(self, tmpdir):
        # Given
        context = CX(tmpdir)
        self.tmpdir = str(tmpdir)
        hp_repository = HyperparamsOnDiskRepository(cache_folder=self.tmpdir)
        n_epochs = 2
        n_trials = 4
        auto_ml = AutoML(
            pipeline=Pipeline([
                MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                    'multiply_by': FixedHyperparameter(2)
                })),
                NumpyReshape(new_shape=(-1, 1)),
                FitTransformCounterLoggingStep()
            ]),
            hyperparams_optimizer=RandomSearchSampler(),
            validation_splitter=ValidationSplitter(0.20),
            scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
            n_trials=n_trials,
            refit_best_trial=True,
            epochs=n_epochs,
            hyperparams_repository=hp_repository,
            continue_loop_on_error=False
        )

        # When
        data_container = DACT(
            data_inputs=np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            expected_outputs=np.array([10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0])
        )
        auto_ml.handle_fit(data_container, context)

        # Then
        file_paths = [os.path.join(
            hp_repository.cache_folder, f"_{DEFAULT_PROJECT}", f"_{DEFAULT_CLIENT}", "_0", f"_{i}", "_0", "log.txt"
        ) for i in range(n_trials)]
        assert len(file_paths) == n_trials

        for f in file_paths:
            assert os.path.exists(f)

        for f in file_paths:
            with open(f, 'r') as f:
                log = f.readlines()
                assert len(log) == 16, log


def test_automl_context_has_loc():
    cx = AutoMLContext.from_context()

    assert cx.loc is not None


def test_automl_context_pushes_loc_attr():
    cx = AutoMLContext.from_context()

    c1 = cx.push_attr(ProjectDataclass(DEFAULT_PROJECT))

    assert cx.loc == ScopedLocation()
    assert c1.loc == ScopedLocation(DEFAULT_PROJECT)


def test_automl_context_loc_pushes_identifier():
    cx = AutoMLContext.from_context()

    c1 = cx.push_attr(ProjectDataclass(DEFAULT_PROJECT))

    assert cx.get_identifier() == NEURAXLE_LOGGER_NAME
    assert c1.get_identifier() == f"{NEURAXLE_LOGGER_NAME}.{DEFAULT_PROJECT}"


def test_root_neuraxle_logger_logs_to_string():
    nxl: NeuraxleLogger = NeuraxleLogger.from_identifier(CX().get_identifier())

    nxl.info("This is a test.")

    assert "This is a test." in nxl.get_scoped_string_history()


def test_automl_neuraxle_logger_logs_to_repo_file(tmpdir):
    cx: AutoMLContext = AutoMLContext.from_context(repo=HyperparamsOnDiskRepository(cache_folder=tmpdir))

    cx.flow.log_status(TrialStatus.RUNNING)
    cx.flow.log_end(TrialStatus.ABORTED)

    log_file_path_at_loc = cx.repo.get_scoped_logger_path(cx.loc)
    assert os.path.exists(log_file_path_at_loc)
    log1 = cx.read_scoped_log()
    with open(log_file_path_at_loc, 'r') as _file:
        log2 = _file.read()
    assert log1 == log2
    assert str(TrialStatus.RUNNING) in log1
    assert str(TrialStatus.ABORTED) in log1


def test_sub_root_neuraxle_loggers_logs_to_string():
    str_r = "Testing root."
    str_a = "Testing a."
    str_b = "Testing b."
    cx = CX()
    nxl_r: NeuraxleLogger = NeuraxleLogger.from_identifier(
        cx.get_identifier())
    nxl_a: NeuraxleLogger = NeuraxleLogger.from_identifier(
        cx.push(Identity(name="a")).get_identifier())
    nxl_b: NeuraxleLogger = NeuraxleLogger.from_identifier(
        cx.push(Identity(name="b")).get_identifier())

    nxl_r.info(str_r)
    nxl_a.info(str_a)
    nxl_b.info(str_b)

    assert str_r in nxl_r.get_scoped_string_history()
    assert str_r not in nxl_a.get_scoped_string_history()
    assert str_r not in nxl_b.get_scoped_string_history()

    assert str_a in nxl_r.get_scoped_string_history()
    assert str_a in nxl_a.get_scoped_string_history()
    assert str_a not in nxl_b.get_scoped_string_history()

    assert str_b in nxl_r.get_scoped_string_history()
    assert str_b not in nxl_a.get_scoped_string_history()
    assert str_b in nxl_b.get_scoped_string_history()


def test_auto_ml_context_services_names():
    cx = AutoMLContext.from_context()

    names: Set[str] = set(cx.getattr("name").to_flat_dict().values())
    assert names == set([
        'AutoMLContext',
        'Flow',
        'VanillaHyperparamsRepository',
        'ScopedLocation',
        'ContextLock'
    ])


def test_automl_context_repo_service_config():
    cx = AutoMLContext.from_context()

    cx.repo.set_config({"some_key": "some_value"})

    assert dict(cx.get_config().to_flat_dict()) == {
        'VanillaHyperparamsRepository__some_key': 'some_value'
    }
    assert cx.has_service("HyperparamsRepository")
    assert cx.get_service("HyperparamsRepository").__class__.__name__ == "VanillaHyperparamsRepository"



"""
Tests for NumPy Steps
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import numpy as np
from neuraxle.steps.numpy import (NumpyConcatenateInnerFeatures,
                                  NumpyFlattenDatum, NumpyShapePrinter,
                                  NumpyTranspose)


def test_flatten_datum():
    flat = NumpyFlattenDatum()
    data = np.random.random((10, 4, 5, 2))  # 4D array (could be ND with N>=2).
    expected_data = np.copy(data).reshape(10, 4 * 5 * 2)  # 2D array.

    flat, received_data = flat.fit_transform(data)

    assert (received_data == expected_data).all()


def test_concat_features():
    concat = NumpyConcatenateInnerFeatures()
    # ND arrays
    data1 = np.random.random((10, 4, 5, 2))
    data2 = np.random.random((10, 4, 5, 10))
    expected_all_data = np.concatenate([data1, data2], axis=-1)

    concat, received_all_data = concat.fit_transform([data1, data2])

    assert tuple(received_all_data.shape) == tuple(expected_all_data.shape)
    assert (received_all_data == expected_all_data).all()


def test_numpy_transpose():
    tr = NumpyTranspose()
    data = np.random.random((10, 7))
    expected_data = np.copy(data).transpose()

    tr, received_data = tr.fit_transform(data)

    assert (received_data == expected_data).all()


def test_numpy_shape_printer():
    pr = NumpyShapePrinter()
    pr.fit_transform(np.ones((10, 11)))



from typing import Any, Tuple

from neuraxle.base import BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import (HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.pipeline import Pipeline
from neuraxle.steps.output_handlers import InputAndOutputTransformerMixin
from py._path.local import LocalPath


class MultiplyBy2OutputTransformer(InputAndOutputTransformerMixin, BaseTransformer):
    def __init__(
            self,
            hyperparams: HyperparameterSamples = None,
            hyperparams_space: HyperparameterSpace = None,
            name: str = None
    ):
        BaseTransformer.__init__(self, hyperparams, hyperparams_space, name)
        InputAndOutputTransformerMixin.__init__(self)

    def transform(self, data_inputs) -> Tuple[Any, Any]:
        dis, eos = data_inputs

        new_dis = []
        new_eos = []
        for di, eo in zip(dis, eos):
            new_dis.append(di * 2)
            new_eos.append(eo * 2)

        return new_dis, new_eos


def test_output_transformer_should_zip_data_input_and_expected_output_in_the_transformed_output(tmpdir: LocalPath):
    pipeline = Pipeline([
        MultiplyBy2OutputTransformer()
    ])

    pipeline, new_data_container = pipeline.handle_fit_transform(
        DACT(data_inputs=[1, 2, 3], ids=[0, 1, 2], expected_outputs=[2, 3, 4]),
        CX(tmpdir)
    )

    assert new_data_container.data_inputs == [2, 4, 6]
    assert new_data_container.expected_outputs == [4, 6, 8]



import numpy as np
import pytest

from neuraxle.base import ExecutionMode
from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import EpochRepeater
from neuraxle.steps.misc import TapeCallbackFunction, FitTransformCallbackStep
from testing.steps.neuraxle_test_case import NeuraxleTestCase

DATA_INPUTS = np.array(range(10))
EXPECTED_OUTPUTS = np.array(range(10, 20))

callback_fit = TapeCallbackFunction()
callback_transform = TapeCallbackFunction()
EPOCHS = 2


@pytest.mark.parametrize("test_case", [
    NeuraxleTestCase(
        pipeline=Pipeline([
            EpochRepeater(FitTransformCallbackStep(callback_transform, callback_fit), epochs=EPOCHS)
        ]),
        callbacks=[callback_fit, callback_transform],
        expected_callbacks_data=[
            [(DATA_INPUTS, EXPECTED_OUTPUTS), (DATA_INPUTS, EXPECTED_OUTPUTS)],
            [DATA_INPUTS]
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=Pipeline([
            EpochRepeater(FitTransformCallbackStep(callback_transform, callback_fit), epochs=EPOCHS)
        ]),
        callbacks=[callback_fit, callback_transform],
        expected_callbacks_data=[
            [],
            [DATA_INPUTS]
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=Pipeline([
            EpochRepeater(FitTransformCallbackStep(callback_transform, callback_fit), epochs=EPOCHS)
        ]).set_train(False),
        callbacks=[callback_fit, callback_transform],
        expected_callbacks_data=[
            [],
            [DATA_INPUTS]
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=Pipeline([
            EpochRepeater(FitTransformCallbackStep(callback_transform, callback_fit), epochs=EPOCHS)
        ]),
        callbacks=[callback_fit, callback_transform],
        expected_callbacks_data=[
            [(DATA_INPUTS, EXPECTED_OUTPUTS), (DATA_INPUTS, EXPECTED_OUTPUTS)],
            []
        ],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        execution_mode=ExecutionMode.FIT
    )
])
def test_epoch_repeater(test_case):
    processed_outputs = test_case.execute()

    test_case.assert_expected_processed_outputs(processed_outputs)
    test_case.assert_callback_data_is_as_expected()



import copy
import datetime
import json

import pytest
from neuraxle.base import Identity, TrialStatus
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.logging.logging import LOGGING_DATETIME_STR_FORMAT
from neuraxle.metaopt.data.aggregates import (MetricResults, Round, Trial,
                                              TrialSplit)
from neuraxle.metaopt.data.vanilla import (AutoMLContext, ClientDataclass,
                                           HyperparamsRepository,
                                           MetricResultsDataclass,
                                           RoundDataclass, ScopedLocation,
                                           TrialDataclass,
                                           VanillaHyperparamsRepository,
                                           from_json, to_json)
from testing.metaopt.test_automl_dataclasses import (SOME_FULL_SCOPED_LOCATION,
                                                   SOME_ROOT_DATACLASS,
                                                   SOME_TRIAL_DATACLASS)

SOME_OTHER_METRIC_NAME = 'MSE'

EXPECTED_ERROR_TRACEBACK = 'NoneType: None\n'

EXPECTED_METRIC_RESULTS = {
    SOME_OTHER_METRIC_NAME: {
        '__type__': MetricResultsDataclass.__name__,
        'metric_name': SOME_OTHER_METRIC_NAME,
        'validation_values': [0.5, 0.7, 0.4],
        'train_values': [0.45, 0.6, 0.4],
        'higher_score_is_better': False,
    }
}


class TestTrials:
    def setup(self):
        self.hp: HyperparameterSamples = HyperparameterSamples({'a': 2})
        round_loc = SOME_FULL_SCOPED_LOCATION[:RoundDataclass]
        self.cx: AutoMLContext = AutoMLContext.from_context().with_loc(round_loc)
        self.trial: Trial = Trial(
            _dataclass=copy.deepcopy(SOME_TRIAL_DATACLASS),
            context=self.cx,
            is_deep=True,
        )
        self.cx.repo.save(
            copy.deepcopy(SOME_ROOT_DATACLASS), ScopedLocation(), deep=True)

    def test_trial_should_have_end_time_later_than_start_time(self):
        with self.trial.new_validation_split(False) as trial_split:
            trial_split._set_success()
            pass

        assert isinstance(trial_split._dataclass.created_time, datetime.datetime)
        assert isinstance(trial_split._dataclass.start_time, datetime.datetime)
        assert isinstance(trial_split._dataclass.end_time, datetime.datetime)
        assert trial_split._dataclass.created_time <= trial_split._dataclass.start_time
        assert trial_split._dataclass.start_time < trial_split._dataclass.end_time

    def test_trial_should_create_new_split(self):
        with self.trial.new_validation_split(False) as trial_split:
            # trial_split.set_success()
            pass

        assert self.trial._validation_splits[-1] == trial_split

    def test_trial_split_is_new_best_score_should_return_true_with_one_score(self):
        with self.trial.new_validation_split(False) as trial_split:
            trial_split: TrialSplit = trial_split.with_n_epochs(1)

            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(0.45)
                metric.add_valid_result(0.5)

                assert metric.is_new_best_score()

    def test_trial_split_is_new_best_score_should_return_false_with_not_a_new_best_score(self):
        with self.trial.new_validation_split(False) as trial_split:
            trial_split: TrialSplit = trial_split
            trial_split.with_n_epochs(2)

            trial_split.next_epoch()
            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(0.45)
                metric.add_valid_result(0.5)

            trial_split.next_epoch()
            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(0.6)
                metric.add_valid_result(0.7)

                assert not metric.is_new_best_score()

    def test_success_trial_split_to_json(self):
        trial_split = self._given_success_trial_validation_split(self.trial)
        trial_json = to_json(trial_split._dataclass)

        self._then_success_trial_split_json_is_valid(json.loads(trial_json))

    def _then_success_trial_split_json_is_valid(self, trial_json):
        assert trial_json['status'] == TrialStatus.SUCCESS.value

        assert MetricResultsDataclass.from_dict(
            trial_json['metric_results'][SOME_OTHER_METRIC_NAME]
        ) == MetricResultsDataclass.from_dict(
            EXPECTED_METRIC_RESULTS[SOME_OTHER_METRIC_NAME])

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], LOGGING_DATETIME_STR_FORMAT)
        created_time = datetime.datetime.strptime(
            trial_json['created_time'], LOGGING_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], LOGGING_DATETIME_STR_FORMAT)
        assert created_time <= start_time
        assert created_time < end_time
        assert start_time < end_time

        return True

    def test_success_trial_to_json(self):
        self._given_success_trial_validation_split(self.trial)

        trial_json = json.loads(to_json(self.trial._dataclass))

        assert trial_json['status'] == TrialStatus.SUCCESS.value

        assert self._then_success_trial_split_json_is_valid(trial_json['validation_splits'][-1])

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], LOGGING_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], LOGGING_DATETIME_STR_FORMAT)

        assert start_time < end_time

    def test_success_trial_get_validation_score(self):
        self._given_success_trial_validation_split(self.trial, best_score=0.3)

        validation_score = self.trial.get_avg_validation_score(SOME_OTHER_METRIC_NAME)

        assert validation_score == 0.3

    def test_success_trial_multiple_splits_should_average_the_scores(self):
        self._given_success_trial_validation_split(self.trial, best_score=0.3)
        self._given_success_trial_validation_split(self.trial, best_score=0.1)

        validation_score = self.trial.get_avg_validation_score(SOME_OTHER_METRIC_NAME)

        assert validation_score == 0.2

    def test_trial_with_failed_split_should_only_average_successful_splits(self):
        self._given_success_trial_validation_split(self.trial, best_score=0.03)
        self._given_success_trial_validation_split(self.trial, best_score=0.01)
        self._given_failed_trial_split_that_continues_on_error(self.trial)

        validation_score = self.trial.get_avg_validation_score(SOME_OTHER_METRIC_NAME)

        assert validation_score == 0.02

    def _given_success_trial_validation_split(self, trial: Trial, best_score=0.4):
        with trial.new_validation_split(False) as trial_split:
            trial_split: TrialSplit = trial_split.with_n_epochs(3)

            trial_split.next_epoch()
            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(0.45)
                metric.add_valid_result(0.5)

            trial_split.next_epoch()
            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(0.6)
                metric.add_valid_result(0.7)

            trial_split.next_epoch()
            with trial_split.managed_metric(
                SOME_OTHER_METRIC_NAME, higher_score_is_better=False
            ) as metric:
                metric: MetricResults = metric

                metric.add_train_result(best_score)
                metric.add_valid_result(best_score)

            return trial_split

    def test_failure_trial_split_to_json(self):
        trial_split = self._given_failed_trial_split_that_continues_on_error(self.trial)

        trial_json = to_json(trial_split._dataclass)

        self._then_failed_validation_split_json_is_valid(json.loads(trial_json), trial_split)

    def _then_failed_validation_split_json_is_valid(self, trial_json, trial_split):
        assert trial_json['status'] == TrialStatus.FAILED.value
        # assert trial_json['error'] == str(trial_split.error)
        # assert trial_json['error_traceback'] == EXPECTED_ERROR_TRACEBACK

        assert MetricResultsDataclass.from_dict(
            trial_json['metric_results'][SOME_OTHER_METRIC_NAME]
        ) == MetricResultsDataclass.from_dict(
            EXPECTED_METRIC_RESULTS[SOME_OTHER_METRIC_NAME])

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], LOGGING_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], LOGGING_DATETIME_STR_FORMAT)
        assert start_time < end_time
        return True

    def test_failure_trial_to_json(self):
        trial_split = self._given_failed_trial_split_that_continues_on_error(self.trial)

        trial_json = json.loads(to_json(self.trial._dataclass))

        # assert trial_json['error'] == str(trial_split.error)
        # assert trial_json['error_traceback'] == EXPECTED_ERROR_TRACEBACK
        assert self._then_failed_validation_split_json_is_valid(
            trial_json['validation_splits'][-1], trial_split)

        created_time = datetime.datetime.strptime(
            trial_json['created_time'], LOGGING_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], LOGGING_DATETIME_STR_FORMAT)

        assert created_time < end_time

    def _given_failed_trial_split_that_continues_on_error(self, trial: Trial):
        """
        This method isn't supposed to raise anything.
        """
        try:
            with trial.new_validation_split(continue_loop_on_error=True) as trial_split:
                trial_split: TrialSplit = trial_split.with_n_epochs(3)

                trial_split.next_epoch()
                with trial_split.managed_metric(
                    SOME_OTHER_METRIC_NAME, higher_score_is_better=False
                ) as metric:
                    metric: MetricResults = metric

                    metric.add_train_result(0.45)
                    metric.add_valid_result(0.5)

                trial_split.next_epoch()
                with trial_split.managed_metric(
                    SOME_OTHER_METRIC_NAME, higher_score_is_better=False
                ) as metric:
                    metric: MetricResults = metric

                    metric.add_train_result(0.6)
                    metric.add_valid_result(0.7)

                trial_split.next_epoch()
                with trial_split.managed_metric(
                    SOME_OTHER_METRIC_NAME, higher_score_is_better=False
                ) as metric:
                    metric: MetricResults = metric

                    metric.add_train_result(0.4)
                    metric.add_valid_result(0.4)

                    raise IndexError('index error')
            return trial._validation_splits[-1]
        except IndexError as e:
            assert False, (
                f"This should not happen: IndexError should be handled by the "
                f"`trial.new_validation_split(continue_loop_on_error=True)` "
                f"context manager... Here is the error: `{e}`.")

    def test_trials_get_best_hyperparams_should_return_hyperparams_of_best_trial(self):
        # Given
        trial_1 = self.trial
        trial_split_1 = self._given_success_trial_validation_split(trial_1, best_score=0.2)

        hp_trial_2 = HyperparameterSamples({'b': 3})
        trial_2 = Trial(
            _dataclass=TrialDataclass(
                trial_number=1,
                hyperparams=hp_trial_2,
            ).start(),
            context=self.cx,
            is_deep=True)
        self.cx.repo.save(trial_2._dataclass, SOME_FULL_SCOPED_LOCATION[:RoundDataclass].with_id(1))
        trial__split_2 = self._given_success_trial_validation_split(trial_2, best_score=0.1)
        self.cx.repo.save(trial_2._set_success()._dataclass, SOME_FULL_SCOPED_LOCATION[:RoundDataclass].with_id(1))

        trials = Round(
            RoundDataclass(trials=[trial_1._dataclass, trial_2._dataclass]),
            context=self.cx.with_loc(SOME_FULL_SCOPED_LOCATION[:ClientDataclass]),
            is_deep=True)

        # When
        best_hyperparams = trials.get_best_hyperparams(SOME_OTHER_METRIC_NAME)

        # Then
        assert best_hyperparams == hp_trial_2
        assert trial_2.is_success()



import numpy as np
from sklearn.cluster import KMeans
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.linear_model import Ridge

from neuraxle.base import Identity
from neuraxle.hyperparams.distributions import RandInt, LogUniform, Boolean
from neuraxle.hyperparams.space import HyperparameterSamples, HyperparameterSpace
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import NumpyTranspose, NumpyConcatenateInnerFeatures
from neuraxle.steps.sklearn import SKLearnWrapper
from neuraxle.union import FeatureUnion, ModelStacking, ZipFeatures


def test_feature_union_should_transform_with_concatenate_inner_features():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=NumpyConcatenateInnerFeatures())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(1, 20))

    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, np.concatenate([data_inputs, data_inputs], axis=-1))


def test_feature_union_should_fit_transform_with_concatenate_inner_features():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=NumpyConcatenateInnerFeatures())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(2, 20))
    expected_outputs = None

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, np.concatenate([data_inputs, data_inputs], axis=-1))


def test_feature_union_should_transform_with_numpy_transpose():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=NumpyTranspose())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(2, 20))

    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, np.array([data_inputs, data_inputs]).transpose())


def test_feature_union_should_fit_transform_with_numpy_transpose():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=NumpyTranspose())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(2, 20))
    expected_outputs = None

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, np.array([data_inputs, data_inputs]).transpose())

def test_feature_union_should_transform_with_zip_features():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=ZipFeatures())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(2, 20))

    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, np.stack([data_inputs, data_inputs], axis=1))


def test_feature_union_should_fit_transform_with_zip_features():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=ZipFeatures())
    ])
    data_inputs = np.random.randint(low=0, high=100, size=(2, 20))
    expected_outputs = None

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, np.stack([data_inputs, data_inputs], axis=1))


def test_feature_union_should_apply_to_self_and_sub_steps():
    p = Pipeline([
        FeatureUnion([
            Identity(),
            Identity(),
        ], joiner=NumpyTranspose())
    ])

    p.apply(lambda step: step._set_hyperparams(HyperparameterSamples({'applied': True})))

    assert p.hyperparams['applied']
    assert p['FeatureUnion'].hyperparams['applied']
    assert p['FeatureUnion'][0].hyperparams['applied']
    assert p['FeatureUnion'][1].hyperparams['applied']
    assert p['FeatureUnion'][2].hyperparams['applied']


def test_model_stacking_fit_transform():
    model_stacking = Pipeline([
        ModelStacking([
            SKLearnWrapper(
                GradientBoostingRegressor(),
                HyperparameterSpace({
                    "n_estimators": RandInt(50, 600), "max_depth": RandInt(1, 10),
                    "learning_rate": LogUniform(0.07, 0.7)
                })
            ),
            SKLearnWrapper(
                KMeans(),
                HyperparameterSpace({
                    "n_clusters": RandInt(5, 10)
                })
            ),
        ],
            joiner=NumpyTranspose(),
            judge=SKLearnWrapper(
                Ridge(),
                HyperparameterSpace({
                    "alpha": LogUniform(0.7, 1.4),
                    "fit_intercept": Boolean()
                })
            ),
        )
    ])
    expected_outputs_shape = (379, 1)
    data_inputs_shape = (379, 13)
    data_inputs = _create_data(data_inputs_shape)
    expected_outputs = _create_data(expected_outputs_shape)

    model_stacking, outputs = model_stacking.fit_transform(data_inputs, expected_outputs)

    assert outputs.shape == expected_outputs_shape


def test_model_stacking_transform():
    model_stacking = Pipeline([
        ModelStacking([
            SKLearnWrapper(
                GradientBoostingRegressor(),
                HyperparameterSpace({
                    "n_estimators": RandInt(50, 600), "max_depth": RandInt(1, 10),
                    "learning_rate": LogUniform(0.07, 0.7)
                })
            ),
            SKLearnWrapper(
                KMeans(),
                HyperparameterSpace({
                    "n_clusters": RandInt(5, 10)
                })
            ),
        ],
            joiner=NumpyTranspose(),
            judge=SKLearnWrapper(
                Ridge(),
                HyperparameterSpace({
                    "alpha": LogUniform(0.7, 1.4),
                    "fit_intercept": Boolean()
                })
            ),
        )
    ])
    expected_outputs_shape = (379, 1)
    data_inputs_shape = (379, 13)
    data_inputs = _create_data(data_inputs_shape)
    expected_outputs = _create_data(expected_outputs_shape)

    model_stacking = model_stacking.fit(data_inputs, expected_outputs)
    outputs = model_stacking.transform(data_inputs)

    assert outputs.shape == expected_outputs_shape


def _create_data(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    return data_inputs



from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import TrainOnlyWrapper
from testing.mocks.step_mocks import SomeSplitStep, SomeStep, SomeTruncableStep


def test_truncable_steps_should_split_by_type():
    pipeline = Pipeline([
        SomeStep(),
        SomeStep(),
        SomeSplitStep(),
        SomeStep(),
        SomeStep(),
        SomeSplitStep(),
        SomeStep(),
    ])

    sub_pipelines = pipeline.split(SomeSplitStep)

    assert 'SomeStep' in sub_pipelines[0]
    assert 'SomeStep1' in sub_pipelines[0]
    assert 'SomeSplitStep' in sub_pipelines[0]
    assert 'SomeStep2' in sub_pipelines[1]
    assert 'SomeStep3' in sub_pipelines[1]
    assert 'SomeSplitStep1' in sub_pipelines[1]
    assert 'SomeStep4' in sub_pipelines[2]


def test_set_train_should_set_train_to_false():
    pipeline = Pipeline([
        SomeStep(),
        SomeStep(),
        Pipeline([
            SomeStep(),
        ])
    ])

    pipeline.set_train(False)

    assert not pipeline.is_train
    assert not pipeline[0].is_train
    assert not pipeline[1].is_train
    assert not pipeline[2].is_train
    assert not pipeline[2][0].is_train


def test_set_train_should_set_train_to_true():
    pipeline = Pipeline([
        SomeStep(),
        SomeStep(),
        Pipeline([
            SomeStep(),
        ])
    ])

    assert pipeline.is_train
    assert pipeline[0].is_train
    assert pipeline[1].is_train
    assert pipeline[2].is_train
    assert pipeline[2][0].is_train


def test_step_print_str_representation_works_correctly():
    output = str(TrainOnlyWrapper(SomeTruncableStep()))
    assert output == """TrainOnlyWrapper(SomeTruncableStep([
        SomeStepWithHyperparams(name='MockStep'),
        SomeStepWithHyperparams(name='MockStep1'),
    ], name='SomeTruncableStep'), name='TrainOnlyWrapper')"""


def test_step_repr_representation_works_correctly():
    output = repr(TrainOnlyWrapper(SomeTruncableStep()))
    assert output == """TrainOnlyWrapper(SomeTruncableStep([
        SomeStepWithHyperparams(name='MockStep', hyperparameters=HyperparameterSamples([('learning_rate', 0.1),
                       ('l2_weight_reg', 0.001),
                       ('hidden_size', 32),
                       ('num_layers', 3),
                       ('num_lstm_layers', 1),
                       ('use_xavier_init', True),
                       ('use_max_pool_else_avg_pool', True),
                       ('dropout_drop_proba', 0.5),
                       ('momentum', 0.1)])),
        SomeStepWithHyperparams(name='MockStep1', hyperparameters=HyperparameterSamples([('learning_rate', 0.1),
                       ('l2_weight_reg', 0.001),
                       ('hidden_size', 32),
                       ('num_layers', 3),
                       ('num_lstm_layers', 1),
                       ('use_xavier_init', True),
                       ('use_max_pool_else_avg_pool', True),
                       ('dropout_drop_proba', 0.5),
                       ('momentum', 0.1)])),
    ], name='SomeTruncableStep', hyperparameters=HyperparameterSamples([('learning_rate', 0.1),
                       ('l2_weight_reg', 0.001),
                       ('hidden_size', 32),
                       ('num_layers', 3),
                       ('num_lstm_layers', 1),
                       ('use_xavier_init', True),
                       ('use_max_pool_else_avg_pool', True),
                       ('dropout_drop_proba', 0.5),
                       ('momentum', 0.1)])), name='TrainOnlyWrapper', hyperparameters=HyperparameterSamples())"""



"""
Apply recursive operations to a pipeline
===========================================================

This demonstrates how to apply a method to each pipeline step.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import json

from scipy.stats import randint

from neuraxle.base import Identity
from neuraxle.hyperparams.space import RecursiveDict, HyperparameterSamples, HyperparameterSpace
from neuraxle.pipeline import Pipeline


class IdentityWithRvs(Identity):
    def _rvs(self):
        return HyperparameterSamples(self.hyperparams_space.rvs())


def rvs(step) -> RecursiveDict:
    return HyperparameterSamples(step.hyperparams_space.rvs())


def main():
    p = Pipeline([
        IdentityWithRvs().set_hyperparams_space(HyperparameterSpace({
            'a': randint(low=2, high=5)
        })),
        IdentityWithRvs().set_hyperparams_space(HyperparameterSpace({
            'b': randint(low=100, high=400)
        }))
    ])

    samples: HyperparameterSamples = p.apply(rvs)
    print('p.apply(rvs) ==>')
    print(json.dumps(samples, indent=4))

    # or equivalently:

    samples: HyperparameterSamples = p.apply('_rvs')
    print('p.apply(\'_rvs\') ==>')
    print(json.dumps(samples, indent=4))


if __name__ == '__main__':
    main()



import os

import numpy as np
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import Identity, StepWithContext
from neuraxle.pipeline import Pipeline
from neuraxle.steps.misc import FitTransformCallbackStep, TapeCallbackFunction
from neuraxle.steps.output_handlers import OutputTransformerWrapper

PIPELINE_NAME = 'saved_pipeline'

DATA_INPUTS = np.array(range(10, 20))
EXPECTED_OUTPUTS = np.array(range(20, 30))


def test_load_full_dump_from_pipeline_name(tmpdir):
    # Given
    tape_fit_callback_function = TapeCallbackFunction()
    tape_transform_callback_function = TapeCallbackFunction()
    pipeline: StepWithContext = Pipeline([
        ('step_a', Identity()),
        ('step_b', OutputTransformerWrapper(
            FitTransformCallbackStep(tape_fit_callback_function, tape_transform_callback_function)
        ))
    ]).set_name(PIPELINE_NAME).with_context(CX(tmpdir))

    # When
    pipeline, _ = pipeline.fit_transform(DATA_INPUTS, EXPECTED_OUTPUTS)

    step_b_wrapped_step = pipeline.wrapped['step_b'].wrapped
    assert np.array_equal(step_b_wrapped_step.transform_callback_function.data[0], EXPECTED_OUTPUTS)
    assert np.array_equal(step_b_wrapped_step.fit_callback_function.data[0][0], EXPECTED_OUTPUTS)
    assert np.array_equal(step_b_wrapped_step.fit_callback_function.data[0][1], [None] * len(EXPECTED_OUTPUTS))

    pipeline.save(CX(tmpdir), full_dump=True)

    # Then
    loaded_pipeline = CX(tmpdir).load(PIPELINE_NAME)

    assert isinstance(loaded_pipeline, Pipeline)
    assert isinstance(loaded_pipeline['step_a'], Identity)
    assert isinstance(loaded_pipeline['step_b'], OutputTransformerWrapper)

    loaded_step_b_wrapped_step = loaded_pipeline['step_b'].wrapped
    assert np.array_equal(loaded_step_b_wrapped_step.transform_callback_function.data[0], EXPECTED_OUTPUTS)
    assert np.array_equal(loaded_step_b_wrapped_step.fit_callback_function.data[0][0], EXPECTED_OUTPUTS)
    assert np.array_equal(loaded_step_b_wrapped_step.fit_callback_function.data[0][1], [None] * len(EXPECTED_OUTPUTS))


def test_load_full_dump_from_path(tmpdir):
    # Given
    tape_fit_callback_function = TapeCallbackFunction()
    tape_transform_callback_function = TapeCallbackFunction()
    pipeline = Pipeline([
        ('step_a', Identity()),
        ('step_b', OutputTransformerWrapper(
            FitTransformCallbackStep(tape_fit_callback_function, tape_transform_callback_function)
        ))
    ]).set_name(PIPELINE_NAME).with_context(CX(tmpdir))

    # When
    pipeline, _ = pipeline.fit_transform(DATA_INPUTS, EXPECTED_OUTPUTS)
    pipeline.save(CX(tmpdir), full_dump=True)

    # Then
    loaded_pipeline = CX(tmpdir).load(os.path.join(PIPELINE_NAME, 'step_b'))

    assert isinstance(loaded_pipeline, OutputTransformerWrapper)
    loaded_step_b_wrapped_step = loaded_pipeline.wrapped
    assert np.array_equal(loaded_step_b_wrapped_step.transform_callback_function.data[0], EXPECTED_OUTPUTS)
    assert np.array_equal(loaded_step_b_wrapped_step.fit_callback_function.data[0][0], EXPECTED_OUTPUTS)
    assert np.array_equal(loaded_step_b_wrapped_step.fit_callback_function.data[0][1], [None] * len(EXPECTED_OUTPUTS))



"""
Validation
====================================
Classes for hyperparameter tuning, such as random search.

..
   Copyright 2021, Neuraxio Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import copy
import math
import operator
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
from functools import reduce
from typing import Any, List, Optional, Set, Tuple

import numpy as np
from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DIT, EOT, IDT, DACTData
from neuraxle.data_container import DataContainer as DACT
from neuraxle.data_container import TrainDACT, ValidDACT
from neuraxle.hyperparams.distributions import (
    ContinuousHyperparameterDistribution, DiscreteHyperparameterDistribution)
from neuraxle.hyperparams.space import (FlatDict, HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.metaopt.data.aggregates import Round
from neuraxle.metaopt.data.vanilla import BaseHyperparameterOptimizer


class RandomSearchSampler(BaseHyperparameterOptimizer):
    """
    AutoML Hyperparameter Optimizer that randomly samples the space of random variables.
    Please refer to :class:`AutoML` for a usage example.

    .. seealso::
        :class:`Trainer`,
        :class:`HyperparamsRepository`,
    """

    def __init__(self):
        BaseHyperparameterOptimizer.__init__(self)

    def find_next_best_hyperparams(self, round_scope: Round) -> HyperparameterSamples:
        """
        Randomly sample the next hyperparams to try.

        :param round_scope: round scope
        :return: next random hyperparams
        """
        return round_scope.hp_space.rvs()


class GridExplorationSampler(BaseHyperparameterOptimizer):
    """
    This hyperparameter space optimizer is similar to a grid search, however, it does
    try to greedily sample maximally different points in the space to explore it. This space
    optimizer has a fixed pseudorandom exploration method that makes the sampling reproductible.

    When over the expected_n_trials (if sampling too much), the sampler will turn
    to a non-seeded random search.

    It may be good for space exploration before a TPE or for unit tests.

    If the expected_n_trials is not set or set to 0, the sampler will guess its ideal
    sampling count and then switch to random search after that.
    """

    def __init__(self, expected_n_trials: int = 0, seed_i: int = 0):
        BaseHyperparameterOptimizer.__init__(self)
        self.expected_n_trials: int = expected_n_trials
        self._i: int = seed_i

    @staticmethod
    def estimate_ideal_n_trials(hp_space: HyperparameterSpace) -> int:
        _ges: GridExplorationSampler = GridExplorationSampler(expected_n_trials=1)
        _ges._reinitialize_grid(hp_space, [])

        _expected_n_trials = _ges.expected_n_trials
        # Then readjust the expected_n_trials to be a multiple of the number of hyperparameters:
        #     NOTE: TRIED A DOZEN OF POWS SQRT CUMSUMPROD AND SOPHISTICATED OTHER THINGS, AND FOUND THIS BEST ONE:
        _estimated_ideal_n_trials: int = math.ceil(0.29 * _expected_n_trials)
        # TODO: could add more samples to the grid lens to match the counts in _estimated_ideal_n_trials.

        return _estimated_ideal_n_trials

    def _reinitialize_grid(self, hp_space: HyperparameterSpace, previous_trials_hp: List[HyperparameterSamples]) -> HyperparameterSamples:
        """
        Update the grid exploration sampler.

        :param round_scope: round scope
        :return: next random hyperparams
        """
        self._n_sampled = 0
        self.flat_hp_grid_values: OrderedDict[str, List[Any]] = {}
        self.flat_hp_grid_lens: List[int] = []
        # TODO: could make use of a ND array here to keep track of the grid exploration instead of using random too much. And a walk method picking the most L2-distant point, permutated over the last mod 3 samples to walk awkwardly like [mid, begin, fartest, side, other side] in the ND cube, also avoiding same-seen values.
        self._seen_hp_grid_values: Set[Tuple[int]] = set()

        self._generate_grid(hp_space)
        for flat_dict_sample in previous_trials_hp:
            self._reshuffle_grid()

            vals: Tuple[int] = tuple(flat_dict_sample.values())
            self._seen_hp_grid_values.add(vals)

    def find_next_best_hyperparams(self, round_scope: Round) -> HyperparameterSamples:
        """
        Sample the next hyperparams to try.

        :param round_scope: round scope
        :return: next hyperparams
        """
        self._reinitialize_grid(round_scope.hp_space, round_scope.get_all_hyperparams())

        _space_max = reduce(operator.mul, self.flat_hp_grid_lens, 1)

        if self._n_sampled >= max(self.expected_n_trials, _space_max):
            return RandomSearchSampler().find_next_best_hyperparams(round_scope)
        for _ in range(_space_max):
            i_grid_keys: Tuple[int] = tuple(self._gen_keys_for_grid())
            if i_grid_keys in self._seen_hp_grid_values:
                self._reshuffle_grid()
                self._i += 1
            else:
                self._seen_hp_grid_values.add(i_grid_keys)
                break

        flat_result: OrderedDict[str, Any] = self[i_grid_keys]
        return HyperparameterSamples(flat_result)

    def _generate_grid(self, hp_space: HyperparameterSpace):
        """
        Generate the grid of hyperparameters to pick from.

        :param hp_space: hyperparameter space
        """
        # Start with discrete params:
        for hp_name, hp_dist in hp_space.to_flat_dict().items():
            if isinstance(hp_dist, DiscreteHyperparameterDistribution):
                hp_samples: List[Any] = hp_dist.values()

                reordered_hp_samples = self._pseudo_shuffle_list(hp_samples)
                self.flat_hp_grid_values[hp_name] = reordered_hp_samples
                self.flat_hp_grid_lens.append(len(reordered_hp_samples))

        # Then fill the remaining continous params using the expected_n_trials:
        remainder: int = max(3, self.expected_n_trials)
        for hp_name, hp_dist in hp_space.to_flat_dict().items():
            if isinstance(hp_dist, ContinuousHyperparameterDistribution):
                hp_samples: List[Any] = [
                    hp_dist.mean(),
                    hp_dist.min() if hp_dist.is_discrete() else hp_dist._pseudo_min(),
                    hp_dist.max() if hp_dist.is_discrete() else hp_dist._pseudo_max(),
                    hp_dist.mean() + hp_dist.std(),
                    hp_dist.mean() - hp_dist.std(),
                    hp_dist.mean() + hp_dist.std() / 2,
                    hp_dist.mean() - hp_dist.std() / 2,
                    hp_dist.mean() + hp_dist.std() * 1.5,
                    hp_dist.mean() - hp_dist.std() * 1.5,
                    hp_dist.mean() + hp_dist.std() / 4,
                    hp_dist.mean() - hp_dist.std() / 4,
                    hp_dist.mean() + hp_dist.std() * 2.5,
                    hp_dist.mean() - hp_dist.std() * 2.5,
                ]

                def _ensure_unique(value: Any, _unique_set: Set[Any]) -> bool:
                    # remove duplicates such as when (or if) STD is of 0.
                    try:
                        ret = value not in _unique_set
                        _unique_set.add(value)
                        return ret
                    except BaseException:
                        return True
                _unique_set: Set[Any] = set()
                hp_samples: List[Any] = [
                    x for x in hp_samples[:remainder]
                    if x in hp_dist and not (math.isinf(x) or math.isnan(x)) and _ensure_unique(x, _unique_set)
                ]
                self.flat_hp_grid_values[hp_name] = hp_samples
                self.flat_hp_grid_lens.append(len(hp_samples))

        _estimated_ideal_n_trials: int = sum(self.flat_hp_grid_lens)

        if self.expected_n_trials != _estimated_ideal_n_trials:
            warnings.warn(
                f"Warning: changed {self.__class__.__name__}.expected_n_trials from "
                f"{self.expected_n_trials} to {_estimated_ideal_n_trials} due to high amount of trials. "
                f"This may lead to a non-reproducible search using RandomSearch as a fallback past this point.")
            self.expected_n_trials = _estimated_ideal_n_trials

        self._i = 1

    @staticmethod
    def _pseudo_shuffle_list(x: List, seed: int = 0) -> list:
        """
        Shuffle a list to create a pseudo-random order that is interesting.
        """
        x = copy.copy(x)
        for i in reversed(range(len(x))):
            v = x[i]
            if (len(x) + i + seed) % 2 and (i + seed) != 0:
                del x[i]
                x.insert(1 - (seed % 2), v)
        return x

    def _gen_keys_for_grid(self) -> List[int]:
        """
        Generate the keys for the grid.

        :param i: index
        :return: keys
        """
        flat_idx: List[int] = []
        for _len in self.flat_hp_grid_lens:
            flat_idx.append(self._i % _len)
        return flat_idx

    def _reshuffle_grid(self, new_sample: FlatDict = None):
        """
        Reshuffling with pseudo-random seed the hyperparameters' values:
        """
        assert self._i != 0, "Cannot reshuffle the grid when _i is 0. Please generate grid first and increase _i."

        for seed, (k, v) in enumerate(self.flat_hp_grid_values.items()):
            if self._i % len(v) == 0:
                # reshuffling the grid for when it comes to the end of each of its sublist.
                # TODO: make a test for this to ensure that over an infinity of trials that the lists are shuffled evenly, or use a real and repeatable pseudorandom rng generator for it.
                new_v = self._pseudo_shuffle_list(v, seed % (self._i + 1) + self._i % (seed + 1))
                self.flat_hp_grid_values[k] = new_v

                if (seed + self._i / len(v)) % 3 == 0:
                    # reversing the grid at each 3 reshuffles + seed as it may sometimes skip the zero in the disorder.
                    self.flat_hp_grid_values[k] = list(reversed(self.flat_hp_grid_values[k]))
                    self._n_sampled += 1

        self._i += 1

    def __getitem__(self, i_grid_keys: List[int]) -> 'OrderedDict[str, Any]':
        """
        Access the keys for the grid.

        :param i_grid_keys: keys
        :return: hyperparams
        """
        flat_result: OrderedDict[str, Any] = OrderedDict()
        for j, hp_name in enumerate(self.flat_hp_grid_values.keys()):
            flat_result[hp_name] = self.flat_hp_grid_values[hp_name][i_grid_keys[j]]
        return flat_result


FoldsList = List  # A list over folds. Can contain DACTData or even DACTs or Tuples of DACTs.


class BaseValidationSplitter(ABC):
    def split_dact(self, data_container: DACT, context: CX) -> FoldsList[Tuple[TrainDACT, ValidDACT]]:
        """
        Wrap a validation split function with a split data container function.
        A validation split function takes two arguments:  data inputs, and expected outputs.

        :param data_container: data container to split
        :return: a tuple of the train and validation data containers.
        """
        splits: FoldsList[Tuple[TrainDACT, ValidDACT]] = []

        data_folds: FoldsList[Tuple[DIT, EOT, IDT, DIT, EOT, IDT]] = list(zip(*self.split(
            data_container.data_inputs, data_container.ids, data_container.expected_outputs, context
        )))

        # Iterate on folds:
        for (train_di, train_eo, train_ids, valid_di, valid_eo, valid_ids) in data_folds:

            # TODO: use ListDACT?
            train_data_container_split: TrainDACT = TrainDACT(
                ids=train_ids,
                data_inputs=train_di,
                expected_outputs=train_eo
            )

            validation_data_container_split: ValidDACT = ValidDACT(
                ids=valid_ids,
                data_inputs=valid_di,
                expected_outputs=valid_eo
            )

            splits.append((train_data_container_split, validation_data_container_split))

        return splits

    @abstractmethod
    def split(
        self,
        data_inputs: DIT,
        ids: Optional[IDT] = None,
        expected_outputs: Optional[EOT] = None,
        context: Optional[CX] = None
    ) -> Tuple[FoldsList[DIT], FoldsList[EOT], FoldsList[IDT], FoldsList[DIT], FoldsList[EOT], FoldsList[IDT]]:
        """
        Train/Test split data inputs and expected outputs.

        :param data_inputs: data inputs
        :param ids: id associated with each data entry (optional)
        :param expected_outputs: expected outputs (optional)
        :param context: execution context (optional)
        :return: train_di, train_eo, train_ids, valid_di, valid_eo, valid_ids
        """
        pass


class ValidationSplitter(BaseValidationSplitter):
    """
    Create a function that splits data into a training, and a validation set.

    .. code-block:: python

        # create a validation splitter function with 80% train, and 20% validation
        validation_splitter(0.20)


    :param test_size: test size in float
    :return:
    """

    def __init__(self, validation_size: float):
        self.validation_size = validation_size

    def split(
        self,
        data_inputs: DIT,
        ids: Optional[IDT] = None,
        expected_outputs: Optional[EOT] = None,
        context: Optional[CX] = None
    ) -> Tuple[FoldsList[DIT], FoldsList[EOT], FoldsList[IDT], FoldsList[DIT], FoldsList[EOT], FoldsList[IDT]]:

        return tuple([
            # The data goes from `DACTData` to `FoldsList[DACTData]`, as per the a single fold:
            [data] for data in self._full_validation_split(
                data_inputs=data_inputs,
                ids=ids,
                expected_outputs=expected_outputs
            )
        ])

    def _full_validation_split(
        self,
        data_inputs: Optional[DIT] = None,
        ids: Optional[IDT] = None,
        expected_outputs: Optional[EOT] = None
    ) -> Tuple[DIT, EOT, IDT, DIT, EOT, IDT]:
        """
        Split data inputs, and expected outputs into a single training set, and a single validation set.

        :param test_size: test size in float
        :param data_inputs: data inputs to split
        :param ids: ids associated with each data entry
        :param expected_outputs: expected outputs to split
        :return: train_di, train_eo, train_ids, valid_di, valid_eo, valid_ids
        """
        return (
            self._train_split(data_inputs),
            self._train_split(expected_outputs),
            self._train_split(ids),
            self._validation_split(data_inputs),
            self._validation_split(expected_outputs),
            self._validation_split(ids),
        )

    def _train_split(self, data_inputs: DACTData) -> DACTData:
        """
        Split training set.

        :param data_inputs: data inputs to split
        :return: train_data_inputs
        """
        if data_inputs is None:
            return None
        return data_inputs[0:self._get_index_split(data_inputs)]

    def _validation_split(self, data_inputs: DACTData) -> DACTData:
        """
        Split validation set.

        :param data_inputs: data inputs to split
        :return: validation_data_inputs
        """
        if data_inputs is None:
            return None
        return data_inputs[self._get_index_split(data_inputs):]

    def _get_index_split(self, data_inputs: DACTData) -> int:
        if self.validation_size < 0 or self.validation_size > 1:
            raise ValueError('validation_size must be a float in the range [0, 1].')
        return math.floor(len(data_inputs) * (1 - self.validation_size))


class KFoldCrossValidationSplitter(BaseValidationSplitter):
    """
    Create a function that splits data with K-Fold Cross-Validation resampling.

    .. code-block:: python

        # create a kfold cross validation splitter with 2 kfold
        kfold_cross_validation_split(0.20)


    :param k_fold: number of folds.
    :return:
    """

    def __init__(self, k_fold: int):
        BaseValidationSplitter.__init__(self)
        self._k_fold: int = k_fold

    def _get_k_fold(self, dact_data: DACTData = None) -> int:
        return self._k_fold

    def split(
        self,
        data_inputs: DIT,
        ids: Optional[IDT] = None,
        expected_outputs: Optional[EOT] = None,
        context: Optional[CX] = None
    ) -> Tuple[FoldsList[DIT], FoldsList[EOT], FoldsList[IDT], FoldsList[DIT], FoldsList[EOT], FoldsList[IDT]]:

        train_di, valid_di = self._kfold_cv_split(
            data_inputs)

        _n_folds = len(train_di)
        _empty_folds: Tuple[FoldsList] = [[None] * _n_folds] * 2

        train_ids, valid_ids = self._kfold_cv_split(
            ids) or copy.deepcopy(_empty_folds)

        train_eo, valid_eo = self._kfold_cv_split(
            expected_outputs) or copy.deepcopy(_empty_folds)

        return train_di, train_eo, train_ids, \
            valid_di, valid_eo, valid_ids

    def _kfold_cv_split(self, dact_data: DACTData) -> Tuple[FoldsList[DACTData], FoldsList[DACTData]]:
        """
        Split data with K-Fold Cross-Validation splitting.

        :param data_inputs: data inputs
        :param k_fold: number of folds
        :return: a tuple of lists of folds of train_data, and of lists of validation_data, each of length "k_fold".
        """
        if dact_data is None:
            return None

        train_splitted_data: List[DACTData] = []
        valid_splitted_data: List[DACTData] = []

        for fold_i in range(self._get_k_fold(dact_data)):
            train_slice, valid_slice = self._get_train_val_slices_at_fold_i(dact_data, fold_i)

            train_splitted_data.append(train_slice)
            valid_splitted_data.append(valid_slice)

        return train_splitted_data, valid_splitted_data

    def _get_train_val_slices_at_fold_i(self, dact_data: DACTData, fold_i: int) -> Tuple[DACTData, DACTData]:
        step = len(dact_data) / float(self._get_k_fold())
        a = int(step * fold_i)
        b = int(step * (fold_i + 1))
        b = min(b, len(dact_data))

        train_slice: DACTData = self._concat_fold_dact_data(dact_data[:a], dact_data[b:])
        valid_slice: DACTData = dact_data[a:b]  # held-out fold against the training data

        return train_slice, valid_slice

    def _concat_fold_dact_data(self, arr1: DACTData, arr2: DACTData) -> DACTData:
        if isinstance(arr1, (list, tuple)):
            return arr1 + arr2
        else:
            return np.concatenate((arr1, arr2), axis=0)


class AnchoredWalkForwardTimeSeriesCrossValidationSplitter(KFoldCrossValidationSplitter):
    """
    An anchored walk forward cross validation works by performing a forward rolling split.

    All training splits start at the beginning of the time series, and finish time varies.

    For the validation split it, will start after a certain time delay (if padding is set)
    after their corresponding training split.

    Data is expected to be an is a square nd.array of shape [batch_size, total_time_steps, ...].
    It can be N dimensions, such as 3D or more, but the time series axis is currently limited to `axis=1`.
    """

    def __init__(
        self,
        minimum_training_size,
        validation_window_size=None,
        padding_between_training_and_validation=0,
        drop_remainder=False,
    ):
        """
        Create a anchored walk forward time series cross validation object.

        The size of the validation split is defined by `validation_window_size`.
        The difference in start position between two consecutive validation split is also equal to
        `validation_window_size`.

        :param minimum_training_size: size of the smallest training split.
        :param validation_window_size: size of each validation split and also the time step taken between each
            forward roll, by default None. If None : It takes the value `minimum_training_size`.
        :param padding_between_training_and_validation: the size of the padding between the end of the training split
            and the start of the validation split, by default 0.
        :param drop_remainder: drop the last split if the last validation split does not coincide
            with a full validation_window_size, by default False.
        """
        self.minimum_training_size = minimum_training_size
        # If validation_window_size is None, we give the same value as training_window_size.
        self.validation_window_size = validation_window_size or self.minimum_training_size
        self.padding_between_training_and_validation = padding_between_training_and_validation
        self.drop_remainder = drop_remainder
        self._validation_initial_start = self.minimum_training_size + self.padding_between_training_and_validation

    def _get_k_fold(self, dact_data: DACTData = None) -> int:
        if self.drop_remainder:
            _round_func = math.floor
        else:
            _round_func = math.ceil
        k_folds = _round_func(
            (dact_data.shape[1] - self._validation_initial_start) / float(self.validation_window_size)
        )
        return k_folds

    def _get_train_val_slices_at_fold_i(self, dact_data: DACTData, fold_i: int) -> Tuple[DACTData, DACTData]:
        # dact_data of shape [batch_size, total_time_steps, ...].

        # first slice index is always 0 for anchored walk forward cross validation.
        a = self._get_beginning_at_fold_i(fold_i)
        b = int(fold_i * self.validation_window_size + self.minimum_training_size)
        b = min(b, dact_data.shape[1])
        train_slice: DACTData = dact_data[:, a:b]

        x = int(fold_i * self.validation_window_size + self._validation_initial_start)
        y = int(x + self.validation_window_size)
        y = min(y, dact_data.shape[1])
        valid_slice: DACTData = dact_data[:, x:y]  # held-out fold against the training data

        return train_slice, valid_slice

    def _get_beginning_at_fold_i(self, fold_i: int) -> int:
        """
        Get the start time of the training split at the given fold index.
        Here in the anchored splitter, it is always zero. This method is overwritten
        in the non-anchored version of the walk forward ts validation splitter
        """
        return 0


class WalkForwardTimeSeriesCrossValidationSplitter(AnchoredWalkForwardTimeSeriesCrossValidationSplitter):
    """
    Perform a classic walk forward cross validation by performing a forward rolling split.
    As opposed to the AnchoredWalkForwardTimeSeriesCrossValidationSplitter, this class
    has a train split that is always of the same size.

    All the training split have the same `validation_window_size` size. The start time and end time of each training
    split will increase identically toward the end at each forward split. Same principle apply with the validation
    split, where the start and end will increase in the same manner toward the end. Each validation split start after
    a certain time delay (if padding is set) after their corresponding training split.

    Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
    The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.
    """

    def __init__(
        self,
        training_window_size,
        validation_window_size=None,
        padding_between_training_and_validation=0,
        drop_remainder=False
    ):
        """
        Create a classic walk forward time series cross validation object.

        The difference in start position between two consecutive validation split are equal to one
        `validation_window_size`.

        :param training_window_size: the window size of training split.
        :param validation_window_size: the window size of each validation split and also the time step taken between
            each forward roll, by default None. If None : It takes the value `training_window_size`.
        :param padding_between_training_and_validation: the size of the padding between the end of the training split
            and the start of the validation split, by default 0.
        :param drop_remainder: drop the last split if the last validation split does not coincide
            with a full validation_window_size, by default False.
        """
        AnchoredWalkForwardTimeSeriesCrossValidationSplitter.__init__(
            self,
            training_window_size,
            validation_window_size=validation_window_size,
            padding_between_training_and_validation=padding_between_training_and_validation,
            drop_remainder=drop_remainder
        )

    def _get_beginning_at_fold_i(self, fold_i: int) -> int:
        """
        Get the start time of the training split at the given fold index.
        """
        return int(fold_i * self.validation_window_size)



"""
Time-related feature engineering with scikit-learn
======================================================

This notebook introduces different strategies with Neuraxle to leverage time-related features
for a bike sharing demand regression task that is highly dependent on business
cycles (days, weeks, months) and yearly season cycles.

In the process, we introduce how to perform periodic feature engineering using
the :class:`sklearn.preprocessing.SplineTransformer` class and its
`extrapolation="periodic"` option.

Machine Learning Pipelines are built using the :class:`~neuraxle.pipeline.Pipeline` class,
as well as :class:`~neuraxle.steps.column_transformer.ColumnTransformer`
and :class:`~neuraxle.union.FeatureUnion`.


..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import matplotlib.pyplot as plt
# %%
# Data exploration on the Bike Sharing Demand dataset
# ---------------------------------------------------
#
# We start by loading the data from the OpenML repository.
import numpy as np
import pandas as pd
from neuraxle.base import Identity, MetaStep
from neuraxle.pipeline import Pipeline
from neuraxle.steps.column_transformer import ColumnTransformer
from neuraxle.union import FeatureUnion

from sklearn.datasets import fetch_openml
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.kernel_approximation import Nystroem
from sklearn.linear_model import RidgeCV
from sklearn.model_selection import TimeSeriesSplit, cross_validate
from sklearn.preprocessing import (FunctionTransformer, MinMaxScaler,
                                   OneHotEncoder, OrdinalEncoder,
                                   PolynomialFeatures, SplineTransformer)

bike_sharing = fetch_openml("Bike_Sharing_Demand", version=2, as_frame=True)
df = bike_sharing.frame

# %%
# To get a quick understanding of the periodic patterns of the data, let us
# have a look at the average demand per hour during a week.
#
# Note that the week starts on a Sunday, during the weekend. We can clearly
# distinguish the commute patterns in the morning and evenings of the work days
# and the leisure use of the bikes on the weekends with a more spread peak
# demand around the middle of the days:


fig, ax = plt.subplots(figsize=(12, 4))
average_week_demand = df.groupby(["weekday", "hour"]).mean()["count"]
average_week_demand.plot(ax=ax)
_ = ax.set(
    title="Average hourly bike demand during the week",
    xticks=[i * 24 for i in range(7)],
    xticklabels=["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"],
    xlabel="Time of the week",
    ylabel="Number of bike rentals",
)

# %%
#
# The target of the prediction problem is the absolute count of bike rentals on
# a hourly basis:
df["count"].max()

# %% [markdown]
#
# Let us rescale the target variable (number of hourly bike rentals) to predict
# a relative demand so that the mean absolute error is more easily interpreted
# as a fraction of the maximum demand.
#
# .. note::
#
#     The fit method of the models used in this notebook all minimize the
#     mean squared error to estimate the conditional mean instead of the mean
#     absolute error that would fit an estimator of the conditional median.
#
#     When reporting performance measure on the test set in the discussion, we
#     instead choose to focus on the mean absolute error that is more
#     intuitive than the (root) mean squared error. Note, however, that the
#     best models for one metric are also the best for the other in this
#     study.
y = df["count"] / 1000

# %%
fig, ax = plt.subplots(figsize=(12, 4))
y.hist(bins=30, ax=ax)
_ = ax.set(
    xlabel="Fraction of rented fleet demand",
    ylabel="Number of hours",
)

# %%
# The input feature data frame is a time annotated hourly log of variables
# describing the weather conditions. It includes both numerical and categorical
# variables. Note that the time information has already been expanded into
# several complementary columns.
#
X = df.drop("count", axis="columns")
X

# %%
# .. note::
#
#    If the time information was only present as a date or datetime column, we
#    could have expanded it into hour-in-the-day, day-in-the-week,
#    day-in-the-month, month-in-the-year using pandas:
#    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#time-date-components
#
# We now introspect the distribution of the categorical variables, starting
# with `"weather"`:
#
X["weather"].value_counts()

# %%
# Since there are only 3 `"heavy_rain"` events, we cannot use this category to
# train machine learning models with cross validation. Instead, we simplify the
# representation by collapsing those into the `"rain"` category.
#
X["weather"].replace(to_replace="heavy_rain", value="rain", inplace=True)
# %%
X["weather"].value_counts()

# %%
# As expected, the `"season"` variable is well balanced:
#
X["season"].value_counts()

# %%
# Time-based cross-validation
# ---------------------------
#
# Since the dataset is a time-ordered event log (hourly demand), we will use a
# time-sensitive cross-validation splitter to evaluate our demand forecasting
# model as realistically as possible. We use a gap of 2 days between the train
# and test side of the splits. We also limit the training set size to make the
# performance of the CV folds more stable.
#
# 1000 test datapoints should be enough to quantify the performance of the
# model. This represents a bit less than a month and a half of contiguous test
# data:


ts_cv = TimeSeriesSplit(
    n_splits=5,
    gap=48,
    max_train_size=10000,
    test_size=1000,
)

# %%
# Let us manually inspect the various splits to check that the
# `TimeSeriesSplit` works as we expect, starting with the first split:
all_splits = list(ts_cv.split(X, y))
train_0, test_0 = all_splits[0]

# %%
X.iloc[test_0]

# %%
X.iloc[train_0]

# %%
# We now inspect the last split:
train_4, test_4 = all_splits[4]

# %%
X.iloc[test_4]

# %%
X.iloc[train_4]

# %%
# All is well. We are now ready to do some predictive modeling!
#
# Gradient Boosting
# -----------------
#
# Gradient Boosting Regression with decision trees is often flexible enough to
# efficiently handle heteorogenous tabular data with a mix of categorical and
# numerical features as long as the number of samples is large enough.
#
# Here, we do minimal ordinal encoding for the categorical variables and then
# let the model know that it should treat those as categorical variables by
# using a dedicated tree splitting rule. Since we use an ordinal encoder, we
# pass the list of categorical values explicitly to use a logical order when
# encoding the categories as integers instead of the lexicographical order.
# This also has the added benefit of preventing any issue with unknown
# categories when using cross-validation.
#
# The numerical variables need no preprocessing and, for the sake of simplicity,
# we only try the default hyper-parameters for this model:


categorical_columns = [
    "weather",
    "season",
    "holiday",
    "workingday",
]
non_categorical_columns = [i for i in X.columns if i not in categorical_columns]
categories = [
    ["clear", "misty", "rain"],
    ["spring", "summer", "fall", "winter"],
    ["False", "True"],
    ["False", "True"],
]
ordinal_encoder = OrdinalEncoder(categories=categories)


gbrt_pipeline = Pipeline([
    ColumnTransformer([
        (categorical_columns, ordinal_encoder),
        (non_categorical_columns, Identity()),
    ], n_dimension=2),
    HistGradientBoostingRegressor(
        categorical_features=range(4),
    ),
])


# %%
#
# Lets evaluate our gradient boosting model with the mean absolute error of the
# relative demand averaged accross our 5 time-based cross-validation splits:


def evaluate(model, X, y, cv):
    class SilentMetaStep(MetaStep):
        """This class is needed here to disable the sklearn compatibility errors with the getters and setters."""

        def __init__(self):
            super().__init__(wrapped=model)

        def get_params(self, deep=False) -> dict:
            return dict()

        def set_params(self, deep=False) -> dict:
            pass
    silent_model = SilentMetaStep()
    cv_results = cross_validate(
        silent_model,
        X,
        y,
        cv=cv,
        scoring=["neg_mean_absolute_error", "neg_root_mean_squared_error"],
    )
    mae = -cv_results["test_neg_mean_absolute_error"]
    rmse = -cv_results["test_neg_root_mean_squared_error"]
    print(
        f"Evaluation results for {model}:\n\n"
        f"    Mean Absolute Error:     {mae.mean():.3f} +/- {mae.std():.3f}\n"
        f"    Root Mean Squared Error: {rmse.mean():.3f} +/- {rmse.std():.3f}\n"
    )


evaluate(gbrt_pipeline, X, y, ts_cv)


# %%
# This model has an average error around 4 to 5% of the maximum demand. This is
# quite good for a first trial without any hyper-parameter tuning! We just had
# to make the categorical variables explicit. Note that the time related
# features are passed as is, i.e. without processing them. But this is not much
# of a problem for tree-based models as they can learn a non-monotonic
# relationship between ordinal input features and the target.
#
# This is not the case for linear regression models as we will see in the
# following.
#
# Naive linear regression
# -----------------------
#
# As usual for linear models, categorical variables need to be one-hot encoded.
# For consistency, we scale the numerical features to the same 0-1 range using
# class:`sklearn.preprocessing.MinMaxScaler`, although in this case it does not
# impact the results much because they are already on comparable scales:


categorical_one_hot_encoders = [
    (col_name, OneHotEncoder(handle_unknown="ignore", sparse=False))
    for col_name in categorical_columns
]
alphas = np.logspace(-6, 6, 25)
naive_linear_pipeline = Pipeline([
    ColumnTransformer([
        *categorical_one_hot_encoders,
        (non_categorical_columns, MinMaxScaler()),
    ], n_dimension=2),
    RidgeCV(alphas=alphas),
])

evaluate(naive_linear_pipeline, X, y, cv=ts_cv)


# %%
#
# The performance is not good: the average error is around 14% of the maximum
# demand. This is more than three times higher than the average error of the
# gradient boosting model. We can suspect that the naive original encoding
# (merely min-max scaled) of the periodic time-related features might prevent
# the linear regression model to properly leverage the time information: linear
# regression does not automatically model non-monotonic relationships between
# the input features and the target. Non-linear terms have to be engineered in
# the input.
#
# For example, the raw numerical encoding of the `"hour"` feature prevents the
# linear model from recognizing that an increase of hour in the morning from 6
# to 8 should have a strong positive impact on the number of bike rentals while
# an increase of similar magnitude in the evening from 18 to 20 should have a
# strong negative impact on the predicted number of bike rentals.
#
# Time-steps as categories
# ------------------------
#
# Since the time features are encoded in a discrete manner using integers (24
# unique values in the "hours" feature), we could decide to treat those as
# categorical variables using a one-hot encoding and thereby ignore any
# assumption implied by the ordering of the hour values.
#
# Using one-hot encoding for the time features gives the linear model a lot
# more flexibility as we introduce one additional feature per discrete time
# level.

hour_weekday_month = ["hour", "weekday", "month"]
non_time_non_categorical_columns = [i for i in non_categorical_columns if i not in hour_weekday_month]

time_one_hot_encoders = [
    (col_name, OneHotEncoder(handle_unknown="ignore", sparse=False))
    for col_name in hour_weekday_month
]

one_hot_linear_pipeline = Pipeline([
    ColumnTransformer([
        *categorical_one_hot_encoders,
        *time_one_hot_encoders,
        (non_time_non_categorical_columns, MinMaxScaler()),
    ], n_dimension=2),
    RidgeCV(alphas=alphas),
])

evaluate(one_hot_linear_pipeline, X, y, cv=ts_cv)


# %%
# The average error rate of this model is 10% which is much better than using
# the original (ordinal) encoding of the time feature, confirming our intuition
# that the linear regression model benefits from the added flexibility to not
# treat time progression in a monotonic manner.
#
# However, this introduces a very large number of new features. If the time of
# the day was represented in minutes since the start of the day instead of
# hours, one-hot encoding would have introduced 1440 features instead of 24.
# This could cause some significant overfitting. To avoid this we could use
# :func:`sklearn.preprocessing.KBinsDiscretizer` instead to re-bin the number
# of levels of fine-grained ordinal or numerical variables while still
# benefitting from the non-monotonic expressivity advantages of one-hot
# encoding.
#
# Finally, we also observe that one-hot encoding completely ignores the
# ordering of the hour levels while this could be an interesting inductive bias
# to preserve to some level. In the following we try to explore smooth,
# non-monotonic encoding that locally preserves the relative ordering of time
# features.
#
# Trigonometric features
# ----------------------
#
# As a first attempt, we can try to encode each of those periodic features
# using a sine and cosine transformation with the matching period.
#
# Each ordinal time feature is transformed into 2 features that together encode
# equivalent information in a non-monotonic way, and more importantly without
# any jump between the first and the last value of the periodic range.


def sin_transformer(period):
    return FunctionTransformer(lambda x: np.sin(x / period * 2 * np.pi))


def cos_transformer(period):
    return FunctionTransformer(lambda x: np.cos(x / period * 2 * np.pi))


# %%
#
# Let us visualize the effect of this feature expansion on some synthetic hour
# data with a bit of extrapolation beyond hour=23:

hour_df = pd.DataFrame(
    np.arange(26).reshape(-1, 1),
    columns=["hour"],
)
hour_df["hour_sin"] = sin_transformer(24).fit_transform(hour_df)["hour"]
hour_df["hour_cos"] = cos_transformer(24).fit_transform(hour_df)["hour"]
hour_df.plot(x="hour")
_ = plt.title("Trigonometric encoding for the 'hour' feature")

# %%
#
# Let's use a 2D scatter plot with the hours encoded as colors to better see
# how this representation maps the 24 hours of the day to a 2D space, akin to
# some sort of a 24 hour version of an analog clock. Note that the "25th" hour
# is mapped back to the 1st hour because of the periodic nature of the
# sine/cosine representation.
fig, ax = plt.subplots(figsize=(7, 5))
sp = ax.scatter(hour_df["hour_sin"], hour_df["hour_cos"], c=hour_df["hour"])
ax.set(
    xlabel="sin(hour)",
    ylabel="cos(hour)",
)
_ = fig.colorbar(sp)

# %%
#
# We can now build a feature extraction pipeline using this strategy:

cyclic_cossin_transformer = ColumnTransformer([
    *categorical_one_hot_encoders,
    ("month", sin_transformer(12)),
    ("month", cos_transformer(12)),
    ("weekday", sin_transformer(7)),
    ("weekday", cos_transformer(7)),
    ("hour", sin_transformer(24)),
    ("hour", cos_transformer(24)),
    (non_time_non_categorical_columns, MinMaxScaler()),
], n_dimension=2)

cyclic_cossin_linear_pipeline = Pipeline([
    cyclic_cossin_transformer,
    RidgeCV(alphas=alphas),
])

evaluate(cyclic_cossin_linear_pipeline, X, y, cv=ts_cv)


# %%
#
# The performance of our linear regression model with this simple feature
# engineering is a bit better than using the original ordinal time features but
# worse than using the one-hot encoded time features. We will further analyze
# possible reasons for this disappointing outcome at the end of this notebook.
#
# Periodic spline features
# ------------------------
#
# We can try an alternative encoding of the periodic time-related features
# using spline transformations with a large enough number of splines, and as a
# result a larger number of expanded features compared to the sine/cosine
# transformation:


def periodic_spline_transformer(period, n_splines=None, degree=3):
    if n_splines is None:
        n_splines = period
    n_knots = n_splines + 1  # periodic and include_bias is True
    return SplineTransformer(
        degree=degree,
        n_knots=n_knots,
        knots=np.linspace(0, period, n_knots).reshape(n_knots, 1),
        extrapolation="periodic",
        include_bias=True,
    )


# %%
#
# Again, let us visualize the effect of this feature expansion on some
# synthetic hour data with a bit of extrapolation beyond hour=23:
hour_df = pd.DataFrame(
    np.linspace(0, 26, 1000).reshape(-1, 1),
    columns=["hour"],
)
splines = periodic_spline_transformer(24, n_splines=12).fit_transform(hour_df)
splines_df = pd.DataFrame(
    splines,
    columns=[f"spline_{i}" for i in range(splines.shape[1])],
)
pd.concat([hour_df, splines_df], axis="columns").plot(x="hour", cmap=plt.cm.tab20b)
_ = plt.title("Periodic spline-based encoding for the 'hour' feature")


# %%
# Thanks to the use of the `extrapolation="periodic"` parameter, we observe
# that the feature encoding stays smooth when extrapolating beyond midnight.
#
# We can now build a predictive pipeline using this alternative periodic
# feature engineering strategy.
#
# It is possible to use fewer splines than discrete levels for those ordinal
# values. This makes spline-based encoding more efficient than one-hot encoding
# while preserving most of the expressivity:
cyclic_spline_transformer = ColumnTransformer([
    *categorical_one_hot_encoders,
    ("month", periodic_spline_transformer(12, n_splines=6)),
    ("weekday", periodic_spline_transformer(7, n_splines=3)),
    ("hour", periodic_spline_transformer(24, n_splines=12)),
    (non_time_non_categorical_columns, MinMaxScaler()),
], n_dimension=2)
cyclic_spline_linear_pipeline = Pipeline([
    cyclic_spline_transformer,
    RidgeCV(alphas=alphas),
])
evaluate(cyclic_spline_linear_pipeline, X, y, cv=ts_cv)


# %%
# Spline features make it possible for the linear model to successfully
# leverage the periodic time-related features and reduce the error from ~14% to
# ~10% of the maximum demand, which is similar to what we observed with the
# one-hot encoded features.
#
# Qualitative analysis of the impact of features on linear model predictions
# --------------------------------------------------------------------------
#
# Here, we want to visualize the impact of the feature engineering choices on
# the time related shape of the predictions.
#
# To do so we consider an arbitrary time-based split to compare the predictions
# on a range of held out data points.
naive_linear_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
naive_linear_predictions = naive_linear_pipeline.predict(X.iloc[test_0])

one_hot_linear_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
one_hot_linear_predictions = one_hot_linear_pipeline.predict(X.iloc[test_0])

cyclic_cossin_linear_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
cyclic_cossin_linear_predictions = cyclic_cossin_linear_pipeline.predict(X.iloc[test_0])

cyclic_spline_linear_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
cyclic_spline_linear_predictions = cyclic_spline_linear_pipeline.predict(X.iloc[test_0])

# %%
# We visualize those predictions by zooming on the last 96 hours (4 days) of
# the test set to get some qualitative insights:
last_hours = slice(-96, None)
fig, ax = plt.subplots(figsize=(12, 4))
fig.suptitle("Predictions by linear models")
ax.plot(
    y.iloc[test_0].values[last_hours],
    "x-",
    alpha=0.2,
    label="Actual demand",
    color="black",
)
ax.plot(naive_linear_predictions[last_hours], "x-", label="Ordinal time features")
ax.plot(
    cyclic_cossin_linear_predictions[last_hours],
    "x-",
    label="Trigonometric time features",
)
ax.plot(
    cyclic_spline_linear_predictions[last_hours],
    "x-",
    label="Spline-based time features",
)
ax.plot(
    one_hot_linear_predictions[last_hours],
    "x-",
    label="One-hot time features",
)
_ = ax.legend()

# %%
# We can draw the following conclusions from the above plot:
#
# - The **raw ordinal time-related features** are problematic because they do
#   not capture the natural periodicity: we observe a big jump in the
#   predictions at the end of each day when the hour features goes from 23 back
#   to 0. We can expect similar artifacts at the end of each week or each year.
#
# - As expected, the **trigonometric features** (sine and cosine) do not have
#   these discontinuities at midnight, but the linear regression model fails to
#   leverage those features to properly model intra-day variations.
#   Using trigonometric features for higher harmonics or additional
#   trigonometric features for the natural period with different phases could
#   potentially fix this problem.
#
# - the **periodic spline-based features** fix those two problems at once: they
#   give more expressivity to the linear model by making it possible to focus
#   on specific hours thanks to the use of 12 splines. Furthermore the
#   `extrapolation="periodic"` option enforces a smooth representation between
#   `hour=23` and `hour=0`.
#
# - The **one-hot encoded features** behave similarly to the periodic
#   spline-based features but are more spiky: for instance they can better
#   model the morning peak during the week days since this peak lasts shorter
#   than an hour. However, we will see in the following that what can be an
#   advantage for linear models is not necessarily one for more expressive
#   models.

# %%
# We can also compare the number of features extracted by each feature
# engineering pipeline:
naive_linear_pipeline[:-1].transform(X).shape

# %%
one_hot_linear_pipeline[:-1].transform(X).shape

# %%
cyclic_cossin_linear_pipeline[:-1].transform(X).shape

# %%
cyclic_spline_linear_pipeline[:-1].transform(X).shape

# %%
# This confirms that the one-hot encoding and the spline encoding strategies
# create a lot more features for the time representation than the alternatives,
# which in turn gives the downstream linear model more flexibility (degrees of
# freedom) to avoid underfitting.
#
# Finally, we observe that none of the linear models can approximate the true
# bike rentals demand, especially for the peaks that can be very sharp at rush
# hours during the working days but much flatter during the week-ends: the most
# accurate linear models based on splines or one-hot encoding tend to forecast
# peaks of commuting-related bike rentals even on the week-ends and
# under-estimate the commuting-related events during the working days.
#
# These systematic prediction errors reveal a form of under-fitting and can be
# explained by the lack of interactions terms between features, e.g.
# "workingday" and features derived from "hours". This issue will be addressed
# in the following section.

# %%
# Modeling pairwise interactions with splines and polynomial features
# -------------------------------------------------------------------
#
# Linear models do not automatically capture interaction effects between input
# features. It does not help that some features are marginally non-linear as is
# the case with features constructed by `SplineTransformer` (or one-hot
# encoding or binning).
#
# However, it is possible to use the `PolynomialFeatures` class on coarse
# grained spline encoded hours to model the "workingday"/"hours" interaction
# explicitly without introducing too many new variables:


hour_workday_interaction = Pipeline([
    ColumnTransformer([
        ("hour", periodic_spline_transformer(24, n_splines=8)),
        ("workingday", FunctionTransformer(lambda x: x == "True")),
    ], n_dimension=2),
    PolynomialFeatures(degree=2, interaction_only=True, include_bias=False)
])

# %%
# Those features are then combined with the ones already computed in the
# previous spline-base pipeline. We can observe a nice performance improvemnt
# by modeling this pairwise interaction explicitly:

cyclic_spline_interactions_pipeline = Pipeline([
    FeatureUnion([
        cyclic_spline_transformer,
        hour_workday_interaction,
    ]),
    RidgeCV(alphas=alphas),
])
evaluate(cyclic_spline_interactions_pipeline, X, y, cv=ts_cv)

# %%
# Modeling non-linear feature interactions with kernels
# -----------------------------------------------------
#
# The previous analysis highlighted the need to model the interactions between
# `"workingday"` and `"hours"`. Another example of a such a non-linear
# interaction that we would like to model could be the impact of the rain that
# might not be the same during the working days and the week-ends and holidays
# for instance.
#
# To model all such interactions, we could either use a polynomial expansion on
# all marginal features at once, after their spline-based expansion. However,
# this would create a quadratic number of features which can cause overfitting
# and computational tractability issues.
#
# Alternatively, we can use the Nyström method to compute an approximate
# polynomial kernel expansion. Let us try the latter:


cyclic_spline_poly_pipeline = Pipeline([
    cyclic_spline_transformer,
    Nystroem(kernel="poly", degree=2, n_components=300, random_state=0),
    RidgeCV(alphas=alphas),
])
evaluate(cyclic_spline_poly_pipeline, X, y, cv=ts_cv)

# %%
#
# We observe that this model can almost rival the performance of the gradient
# boosted trees with an average error around 6% of the maximum demand.
#
# Note that while the final step of this pipeline is a linear regression model,
# the intermediate steps such as the spline feature extraction and the Nyström
# kernel approximation are highly non-linear. As a result the compound pipeline
# is much more expressive than a simple linear regression model with raw features.
#
# For the sake of completeness, we also evaluate the combination of one-hot
# encoding and kernel approximation:

one_hot_poly_pipeline = Pipeline([
    ColumnTransformer([
        *categorical_one_hot_encoders,
        *time_one_hot_encoders,
        (non_time_non_categorical_columns, Identity())  # passthrough
    ], n_dimension=2),
    Nystroem(kernel="poly", degree=2, n_components=300, random_state=0),
    RidgeCV(alphas=alphas)
])
evaluate(one_hot_poly_pipeline, X, y, cv=ts_cv)


# %%
# While one-hot encoded features were competitive with spline-based features
# when using linear models, this is no longer the case when using a low-rank
# approximation of a non-linear kernel: this can be explained by the fact that
# spline features are smoother and allow the kernel approximation to find a
# more expressive decision function.
#
# Let us now have a qualitative look at the predictions of the kernel models
# and of the gradient boosted trees that should be able to better model
# non-linear interactions between features:
gbrt_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
gbrt_predictions = gbrt_pipeline.predict(X.iloc[test_0])

one_hot_poly_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
one_hot_poly_predictions = one_hot_poly_pipeline.predict(X.iloc[test_0])

cyclic_spline_poly_pipeline.fit(X.iloc[train_0], y.iloc[train_0])
cyclic_spline_poly_predictions = cyclic_spline_poly_pipeline.predict(X.iloc[test_0])

# %%
# Again we zoom on the last 4 days of the test set:

last_hours = slice(-96, None)
fig, ax = plt.subplots(figsize=(12, 4))
fig.suptitle("Predictions by non-linear regression models")
ax.plot(
    y.iloc[test_0].values[last_hours],
    "x-",
    alpha=0.2,
    label="Actual demand",
    color="black",
)
ax.plot(
    gbrt_predictions[last_hours],
    "x-",
    label="Gradient Boosted Trees",
)
ax.plot(
    one_hot_poly_predictions[last_hours],
    "x-",
    label="One-hot + polynomial kernel",
)
ax.plot(
    cyclic_spline_poly_predictions[last_hours],
    "x-",
    label="Splines + polynomial kernel",
)
_ = ax.legend()


# %%
# First, note that trees can naturally model non-linear feature interactions
# since, by default, decision trees are allowed to grow beyond a depth of 2
# levels.
#
# Here, we can observe that the combinations of spline features and non-linear
# kernels works quite well and can almost rival the accuracy of the gradient
# boosting regression trees.
#
# On the contrary, one-hot encoded time features do not perform that well with
# the low rank kernel model. In particular, they significantly over-estimate
# the low demand hours more than the competing models.
#
# We also observe that none of the models can successfully predict some of the
# peak rentals at the rush hours during the working days. It is possible that
# access to additional features would be required to further improve the
# accuracy of the predictions. For instance, it could be useful to have access
# to the geographical repartition of the fleet at any point in time or the
# fraction of bikes that are immobilized because they need servicing.
#
# Let us finally get a more quantative look at the prediction errors of those
# three models using the true vs predicted demand scatter plots:
fig, axes = plt.subplots(ncols=3, figsize=(12, 4), sharey=True)
fig.suptitle("Non-linear regression models")
predictions = [
    one_hot_poly_predictions,
    cyclic_spline_poly_predictions,
    gbrt_predictions,
]
labels = [
    "One hot + polynomial kernel",
    "Splines + polynomial kernel",
    "Gradient Boosted Trees",
]
for ax, pred, label in zip(axes, predictions, labels):
    ax.scatter(y.iloc[test_0].values, pred, alpha=0.3, label=label)
    ax.plot([0, 1], [0, 1], "--", label="Perfect model")
    ax.set(
        xlim=(0, 1),
        ylim=(0, 1),
        xlabel="True demand",
        ylabel="Predicted demand",
    )
    ax.legend()


# %%
# This visualization confirms the conclusions we draw on the previous plot.
#
# All models under-estimate the high demand events (working day rush hours),
# but gradient boosting a bit less so. The low demand events are well predicted
# on average by gradient boosting while the one-hot polynomial regression
# pipeline seems to systematically over-estimate demand in that regime. Overall
# the predictions of the gradient boosted trees are closer to the diagonal than
# for the kernel models.
#
# Concluding remarks
# ------------------
#
# We note that we could have obtained slightly better results for kernel models
# by using more components (higher rank kernel approximation) at the cost of
# longer fit and prediction durations. For large values of `n_components`, the
# performance of the one-hot encoded features would even match the spline
# features.
#
# The `Nystroem` + `RidgeCV` regressor could also have been replaced by
# :class:`~sklearn.neural_network.MLPRegressor` with one or two hidden layers
# and we would have obtained quite similar results.
#
# The dataset we used in this case study is sampled on a hourly basis. However
# cyclic spline-based features could model time-within-day or time-within-week
# very efficiently with finer-grained time resolutions (for instance with
# measurements taken every minute instead of every hours) without introducing
# more features. One-hot encoding time representations would not offer this
# flexibility.
#
# Finally, in this notebook we used `RidgeCV` because it is very efficient from
# a computational point of view. However, it models the target variable as a
# Gaussian random variable with constant variance. For positive regression
# problems, it is likely that using a Poisson or Gamma distribution would make
# more sense. This could be achieved by using
# `GridSearchCV(TweedieRegressor(power=2), param_grid({"alpha": alphas}))`
# instead of `RidgeCV`.
#
# Source
# ------------------
#
# .. note::
#
#     Some of the code in the current file is adapted from the following
#     `scikit-learn documentation example <https://scikit-learn.org/stable/auto_examples/applications/plot_cyclical_feature_engineering.html#sphx-glr-auto-examples-applications-plot-cyclical-feature-engineering-py>`__,
#     and was updated to Neuraxle code for demonstration herein.
#     Here is the license of the original example:
#
#     BSD 3-Clause License
#
#     Copyright (c) 2007-2021 The scikit-learn developers.
#     All rights reserved.
#
#     Redistribution and use in source and binary forms, with or without
#     modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#     AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#     IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#     FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#     DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#     SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#     CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#     OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#     OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#



"""
Easy REST API Model Serving with Neuraxle
================================================

This demonstrates an easy way to deploy your Neuraxle model or pipeline to a REST API.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np
from flask import Flask
from sklearn.cluster import KMeans
from sklearn.datasets import load_boston
from sklearn.decomposition import PCA, FastICA
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

from neuraxle.rest.flask import FlaskRestApiWrapper, JSONDataBodyDecoder, JSONDataResponseEncoder
from neuraxle.pipeline import Pipeline
from neuraxle.steps.sklearn import RidgeModelStacking
from neuraxle.union import AddFeatures


def main():
    boston = load_boston()
    X, y = shuffle(boston.data, boston.target, random_state=13)
    X = X.astype(np.float32)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, shuffle=False)

    pipeline = Pipeline([
        AddFeatures([
            PCA(n_components=2),
            FastICA(n_components=2),
        ]),
        RidgeModelStacking([
            GradientBoostingRegressor(),
            KMeans(),
        ]),
    ])

    print("Fitting on train:")
    pipeline = pipeline.fit(X_train, y_train)
    print("")
    print("Transforming train and test:")
    y_train_predicted = pipeline.transform(X_train)
    y_test_predicted = pipeline.transform(X_test)
    print("")
    print("Evaluating transformed train:")
    score = r2_score(y_train_predicted, y_train)
    print('R2 regression score:', score)
    print("")
    print("Evaluating transformed test:")
    score = r2_score(y_test_predicted, y_test)
    print('R2 regression score:', score)
    print("Deploying the application by routing data to the transform method:")

    class CustomJSONDecoderFor2DArray(JSONDataBodyDecoder):
        """This is a custom JSON decoder class that precedes the pipeline's transformation."""

        def decode(self, data_inputs):
            """
            Transform a JSON list object into an np.array object.

            :param data_inputs: json object
            :return: np array for data inputs
            """
            return np.array(data_inputs)

    class CustomJSONEncoderOfOutputs(JSONDataResponseEncoder):
        """This is a custom JSON response encoder class for converting the pipeline's transformation outputs."""

        def encode(self, data_inputs) -> dict:
            """
            Convert predictions to a dict for creating a JSON Response object.

            :param data_inputs:
            :return:
            """
            return {
                'predictions': list(data_inputs)
            }

    app = FlaskRestApiWrapper(
        json_decoder=CustomJSONDecoderFor2DArray(),
        wrapped=pipeline,
        json_encoder=CustomJSONEncoderOfOutputs()
    ).get_app()

    print("Finally, run the app by uncommenting this next line of code:")

    # app.run(debug=False, port=5000)

    print("You can now call your pipeline over HTTP with a (JSON) REST API.")

    # test_predictictions = requests.post(
    #     url='http://127.0.0.1:5000/',
    #     json=X_test.tolist()
    # )
    # print(test_predictictions)
    # print(test_predictictions.content)

    assert isinstance(app, Flask)

    return app


if __name__ == "__main__":
    main()
